# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from bitfount.federated.transport.protos import (
    messages_pb2 as bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2,
)


class MessageServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PodConnect = channel.unary_unary(
            "/messages.MessageService/PodConnect",
            request_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.PodData.SerializeToString,
            response_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.FromString,
        )
        self.SetupTaskMailboxes = channel.unary_unary(
            "/messages.MessageService/SetupTaskMailboxes",
            request_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountTasks.SerializeToString,
            response_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.CommunicationDetails.FromString,
        )
        self.AcknowledgeMessage = channel.unary_unary(
            "/messages.MessageService/AcknowledgeMessage",
            request_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.Acknowledgement.SerializeToString,
            response_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.FromString,
        )
        self.GetBitfountMessage = channel.unary_unary(
            "/messages.MessageService/GetBitfountMessage",
            request_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.CommunicationDetails.SerializeToString,
            response_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountMessage.FromString,
        )
        self.SendBitfountMessage = channel.unary_unary(
            "/messages.MessageService/SendBitfountMessage",
            request_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountMessage.SerializeToString,
            response_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.FromString,
        )
        self.GetLargeObjectStorage = channel.unary_unary(
            "/messages.MessageService/GetLargeObjectStorage",
            request_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.LargeStorageRequest.SerializeToString,
            response_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BlobStorageData.FromString,
        )


class MessageServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def PodConnect(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SetupTaskMailboxes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AcknowledgeMessage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetBitfountMessage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SendBitfountMessage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetLargeObjectStorage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_MessageServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "PodConnect": grpc.unary_unary_rpc_method_handler(
            servicer.PodConnect,
            request_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.PodData.FromString,
            response_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.SerializeToString,
        ),
        "SetupTaskMailboxes": grpc.unary_unary_rpc_method_handler(
            servicer.SetupTaskMailboxes,
            request_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountTasks.FromString,
            response_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.CommunicationDetails.SerializeToString,
        ),
        "AcknowledgeMessage": grpc.unary_unary_rpc_method_handler(
            servicer.AcknowledgeMessage,
            request_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.Acknowledgement.FromString,
            response_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.SerializeToString,
        ),
        "GetBitfountMessage": grpc.unary_unary_rpc_method_handler(
            servicer.GetBitfountMessage,
            request_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.CommunicationDetails.FromString,
            response_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountMessage.SerializeToString,
        ),
        "SendBitfountMessage": grpc.unary_unary_rpc_method_handler(
            servicer.SendBitfountMessage,
            request_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountMessage.FromString,
            response_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.SerializeToString,
        ),
        "GetLargeObjectStorage": grpc.unary_unary_rpc_method_handler(
            servicer.GetLargeObjectStorage,
            request_deserializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.LargeStorageRequest.FromString,
            response_serializer=bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BlobStorageData.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "messages.MessageService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class MessageService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def PodConnect(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/messages.MessageService/PodConnect",
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.PodData.SerializeToString,
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SetupTaskMailboxes(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/messages.MessageService/SetupTaskMailboxes",
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountTasks.SerializeToString,
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.CommunicationDetails.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AcknowledgeMessage(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/messages.MessageService/AcknowledgeMessage",
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.Acknowledgement.SerializeToString,
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetBitfountMessage(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/messages.MessageService/GetBitfountMessage",
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.CommunicationDetails.SerializeToString,
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SendBitfountMessage(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/messages.MessageService/SendBitfountMessage",
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BitfountMessage.SerializeToString,
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.SuccessResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetLargeObjectStorage(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/messages.MessageService/GetLargeObjectStorage",
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.LargeStorageRequest.SerializeToString,
            bitfount_dot_federated_dot_transport_dot_protos_dot_messages__pb2.BlobStorageData.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



"""RSA Blinding Private Set intersection.

Order of operations:

1. Modeller sends the task request to the worker, containing
    the columns on which the intersection should be performed.
2. Worker generates an RSA key pair, and sends the public key
    to the modeller, keeping the private key to themself.
3. Worker hashes and uses his private key to encrypt each
    row for the relevant columns, and then hashes again the obtained values.

    Modeller receives the public key from the pod. Then he
    chooses random factors (as many as the size of the rows
    in his data), which he encrypts using the public key.
    He then hashes each row of his input and multiplies it
    with the encrypted random factors, which is referred
    to as blinding.
4. Modeller sends his blinded inputs to the worker.
5. Worker receives the blinded inputs from the modeller.
    He first encrypts the blinded values using his private
    key, and then hashes each of the results. He then sends
    the list of blinded hashed values and the list of
    hashed values from step 3 to the modeller.
6. Modeller receives the two lists from the pod.
    He then can compute the private set intersection.

For more details see: https://eprint.iacr.org/2009/491.pdf
"""
import secrets
from typing import Any, List, Optional, Tuple

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
import pandas as pd

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.base import (
    _BaseAlgorithmFactory,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.exceptions import (
    BlindingError,
    OutOfBoundsError,
    PSIMultiTableError,
    UnBlindingError,
)
from bitfount.federated.mixins import _PSIAlgorithmsMixIn
from bitfount.federated.types import _DataLessAlgorithm
from bitfount.federated.utils import hash_set
from bitfount.utils import _powinv, _powmod, delegates


class _ModellerSide(_BaseModellerAlgorithm):
    """Modeller side of the RSABlind algorithm.

    Attributes:
        first_hash_function: The first hash function used in the psi algorithm.
        second_hash_function: The second hash function used in the psi algorithm.
    """

    def __init__(self, **kwargs: Any) -> None:
        self.first_hash_function = hashes.SHA256()
        self.second_hash_function = hashes.SHA512_256()

    def initialise(
        self,
        datasource: BaseSource,
        public_key: RSAPublicKey,
        columns_to_intersect: Optional[List[str]] = None,
        table: Optional[str] = None,
        **kwargs: Any,
    ) -> List[int]:
        """Initialize the psi algorithm on the modeller side.

        More specifically, set the public key parameters,
        load and hash the data, generate random factors, and blind the dataset.

        Args:
            public_key: The public key received from the pod.
            datasource: The datasource on which the psi will be performed.
            columns_to_intersect: Either a list of column names for a
            single-table datasource, or a mapping of table name
            to columns for multi-table datasources.

        Returns:
            The hashed and blinded set.

        Raises:
            PSIColumnsError if you are trying to load columns from more than one table,
                or if the format of `columns_to_intersect` is not supported.
        """
        # Public key consists of `n`, the modulus, and `e`, the encryption key.
        self.n, self.e = self._get_public_key_numbers(public_key)
        # Load the data; `self.data` should be a dataframe.
        if datasource.multi_table and table is None:
            raise PSIMultiTableError(
                "You are trying to perform a PrivateSetIntersection task on a "
                "multitable datasource. Please specify the target table on which "
                "the Private Set Intersection should be performed."
            )
        elif columns_to_intersect and table:
            self.data = datasource.get_data(table)[columns_to_intersect]
        elif columns_to_intersect:
            self.data = datasource.get_data()[columns_to_intersect]
        elif table:
            self.data = datasource.get_data(table)
        else:
            self.data = datasource.get_data()

        # Compute the hashes
        self.hashed_data = hash_set(self.data, self.first_hash_function)
        # Generate random factors for blinding
        self.random_factors = self.generate_random_factors(len(self.data))

        return self.blind_set()

    def _get_public_key_numbers(self, public_key: RSAPublicKey) -> Tuple[int, int]:
        """Get the public key parameters from the RSA public key.

        Returns:
            `n`, the modulus for the public key, and `e`, the public encryption key.
        """
        return public_key.public_numbers().n, public_key.public_numbers().e

    def generate_random_factors(self, n_elements: int) -> List[Tuple[int, int]]:
        """Generate n_elements random factors, then invert and encrypt them.

        Args:
           n_elements: number of random factors to generate.

        Returns:
           A list of tuples, each tuple is composed of the inverse and
           encryption of a random factor.
        """
        random_factors = []
        iteration = 0
        while iteration < n_elements:
            r = secrets.randbelow(self.n)
            try:
                r_inv = _powinv(r, self.n)
            except ValueError:
                # If the random value is not invertible continue and try again
                continue
            r_encrypted = _powmod(r, self.e, self.n)
            iteration += 1
            random_factors.append((r_inv, r_encrypted))
        return random_factors

    def blind(self, record: int, rf_tuple: Tuple[int, int]) -> int:
        """Blind an element using the encryption of a random factor.

        Args:
            record: integer in the range [0, n), where n is the RSA modulus.
            rf_tuple: tuple composed of the inverse and encryption of a random
                factor.

        Returns:
            The blinded version of the element integer x.
        """
        rf_encrypted = rf_tuple[1]
        return (record * rf_encrypted) % self.n

    def unblind(self, integer: int, random_factor_tuple: Tuple[int, int]) -> int:
        """Blind an element using the inverse of a random factor.

        Args:
            integer: integer in the range [0, n), where n is the RSA modulus.
            random_factor_tuple: tuple composed of the inverse and
                encryption of a random factor.

        Returns:
            The unblinded version of the blinded element x.
        """
        rf_inv = random_factor_tuple[0]
        return (integer * rf_inv) % self.n

    def blind_set(self) -> List[int]:
        """Blind a set of elements using the encryption of random factors.

        Returns:
            A list of integers representing the blinded set.
        """
        if len(self.hashed_data) > len(self.random_factors):
            raise BlindingError("There are more elements than random factors")

        blinded_set = []
        for x, rf in zip(self.hashed_data, self.random_factors):
            b = self.blind(x, rf)
            blinded_set.append(b)

        return blinded_set

    def unblind_set(self, dataset: List[int]) -> List[int]:
        """Unblind a set of elements using the inverse of random factors.

        Args:
            dataset: list of integers in the range [0, n), where n is the RSA modulus.

        Returns:
            A list of integers representing the unblinded set X.
        """
        if len(dataset) > len(self.random_factors):
            raise UnBlindingError("There are more elements than random factors")

        unblinded_set = []
        for record, rf_tuple in zip(dataset, self.random_factors):
            unblinded = self.unblind(record, rf_tuple)
            unblinded_set.append(unblinded)
        return unblinded_set

    def intersect(
        self,
        modeller_orig_df: pd.DataFrame,
        blinded_modeller_set: List[int],
        pod_set: List[int],
    ) -> pd.DataFrame:
        """Compute the intersection of the client and server set.

        Args:
            modeller_orig_df: ordered client set.
            blinded_modeller_set: ordered signatures of the client set (unblinded).
            pod_set: signed server set.

        Returns:
            A list of integers representing the intersection of the client and
            server set.
        """
        result = []
        for index in range(len(blinded_modeller_set)):
            if blinded_modeller_set[index] in pod_set:
                result.append(modeller_orig_df.iloc[index])
        return pd.DataFrame(result)

    def run(self, pod_set: List[int], modeller_set: List[int]) -> pd.DataFrame:
        """Runs the modeller side of the algorithm.

        First, it unblinds the modeller set obtained from the worker.
        Then, it hashes the obtained set, and finally performs the intersection.
        """
        unblinded_set = self.unblind_set(modeller_set)
        hashed_modeller_set = hash_set(unblinded_set, self.second_hash_function)
        return self.intersect(self.data, hashed_modeller_set, pod_set)


class _WorkerSide(_BaseWorkerAlgorithm):
    """Worker side of the RSABlind algorithm."""

    def __init__(self, **kwargs: Any) -> None:
        # Set up hash functions algorithms to use
        self.first_hash_function = hashes.SHA256()
        self.second_hash_function = hashes.SHA512_256()
        # Set up private and public key parameters
        self.private_key, self.public_key = _RSAEncryption.generate_key_pair()
        self.n = self.public_key.public_numbers().n
        self.d = self.private_key.private_numbers().d

    def initialise(
        self,
        datasource: BaseSource,
        columns_to_intersect: Optional[List[str]] = None,
        table: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> List[int]:
        """Initialise the worker-side algorithm.

        Args:
            datasource: The datasource of the worker on which the
                intersection will be performed.
            columns_to_intersect: The columns from the worker datasource
                on which to perform the intersection.

        Returns:
            The pod's double hashed set.

        Raises:
            PSIColumnsError if you are trying to load columns from more than one table,
                or if the format of `columns_to_intersect` is not supported.
        """
        if datasource.multi_table and table is None:
            raise PSIMultiTableError(
                "You are trying to perform a PrivateSetIntersection task on a "
                "multitable datasource. Please specify the target table on which "
                "the Private Set Intersection should be performed."
            )
        elif columns_to_intersect and table:
            self.data = datasource.get_data(table)[columns_to_intersect]
        elif columns_to_intersect:
            self.data = datasource.get_data()[columns_to_intersect]
        elif table:
            self.data = datasource.get_data(table)
        else:
            self.data = datasource.get_data()

        hashed_data = hash_set(self.data, self.first_hash_function)
        decrypted_set = self.decrypt_set(hashed_data)
        return hash_set(decrypted_set, self.second_hash_function)

    def run(self, modeller_set: List[int]) -> List[int]:
        """Runs the worker-side of the algorithm."""
        return self.decrypt_set(modeller_set)

    def decrypt(self, item: int) -> int:
        """Sign a single element using the RSA private key.

        Args:
            item: integer in the range [0, n), where n is the RSA modulus.

        Returns:
            The signature of item, computed as item^d modulus n.
        """
        if item < 0 or item >= self.n:
            raise OutOfBoundsError(f"x should be in range [0, {self.n})")
        return _powmod(item, self.d, self.n)

    def decrypt_set(self, data: List[int]) -> List[int]:
        """Decrypt a set of elements using the RSA private key.

        Args:
            data: list of integers in the range [0, n), where n is the RSA modulus.

        Returns:
            A list of integers representing the signatures of the set X.
        """
        return [self.decrypt(item) for item in data]


@delegates()
class ComputeIntersectionRSA(
    _BaseAlgorithmFactory, _PSIAlgorithmsMixIn, _DataLessAlgorithm
):
    """Algorithm for computing the private set intersection with rsa blinding."""

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        super().__init__()

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the SqlQuery algorithm."""
        return _ModellerSide(**kwargs)

    def worker(self, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the SqlQuery algorithm."""
        return _WorkerSide(**kwargs)



"""Tests for the monitor and decorator modules."""
from datetime import datetime, timezone
from itertools import chain
import random
from typing import Dict, Optional, Type, Union
from unittest.mock import Mock, create_autospec

from _pytest.logging import LogCaptureFixture
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
from requests import HTTPError, RequestException

from bitfount import BitfountHub
from bitfount.federated.monitoring import (
    AdditionalMonitorMessageTypes,
    ExistingMonitorModuleError,
    MonitorRecordPrivacy,
    NoMonitorModuleError,
    ProgressCounterDict,
    monitor,
)
from bitfount.federated.monitoring.decorators import task_status
from bitfount.federated.monitoring.monitor import (
    _MONITOR_MODULE_LOCK,
    _get_task_monitor,
    _MonitorModule,
    _MonitorPostContent,
    _set_task_monitor,
    _unset_task_monitor,
    task_config_update,
    task_monitor_context,
    task_status_update,
)
from bitfount.federated.transport.message_service import _BitfountMessageType
from bitfount.hub.types import _MonitorPostJSON
from bitfount.types import _JSONDict
from tests.utils import PytestRequest
from tests.utils.helper import get_warning_logs, unit_test


@fixture
def wrapped_monitor_lock(mocker: MockerFixture) -> Mock:
    """Wraps the monitor lock in a mock so call assertions can be made."""
    mock_lock: Mock = mocker.patch(
        "bitfount.federated.monitoring.monitor._MONITOR_MODULE_LOCK",
        wraps=_MONITOR_MODULE_LOCK,
    )
    mock_lock.__enter__ = mock_lock.acquire
    mock_lock.__exit__ = lambda *_: mock_lock.release()
    return mock_lock


@fixture
def mock_hub() -> Mock:
    """Mock BitfountHub."""
    mock_hub: Mock = create_autospec(BitfountHub, instance=True)
    return mock_hub


@fixture
def task_id() -> str:
    """Task ID."""
    return "a-task-id"


@fixture
def sender_id() -> str:
    """Sender ID."""
    return "sender-mailbox-id"


@fixture(params=(True, False), ids=("recipient_id", "no_recipient_id"))
def recipient_id(request: PytestRequest) -> Optional[str]:
    """Recipient ID or None."""
    incl_recip_id: bool = request.param
    if incl_recip_id:
        return "recipient-mailbox-id"
    else:
        return None


@fixture
def mock_monitor_module() -> Mock:
    """A mock monitor module instance."""
    mock_monitor_module: Mock = create_autospec(_MonitorModule, instance=True)
    return mock_monitor_module


@unit_test
class TestMonitorModuleManagement:
    """Tests for the global monitor module control code."""

    def test__set_task_monitor_sets_monitor(
        self,
        mock_hub: Mock,
        mocker: MockerFixture,
        recipient_id: Optional[str],
        sender_id: str,
        task_id: str,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test global monitor module can be set properly."""
        # Ensure _MONITOR_MODULE is None at beginning
        mocker.patch("bitfount.federated.monitoring.monitor._MONITOR_MODULE", None)

        _set_task_monitor(mock_hub, task_id, sender_id, recipient_id)

        # Check lock was used
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check monitor module has been created with correct task ID
        assert monitor._MONITOR_MODULE is not None
        assert monitor._MONITOR_MODULE.task_id == task_id

    def test__set_task_monitor_fails_if_existing(
        self,
        mock_hub: Mock,
        mocker: MockerFixture,
        recipient_id: Optional[str],
        sender_id: str,
        task_id: str,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test global monitor module can be set properly."""
        # Ensure _MONITOR_MODULE is NOT None at beginning
        mock_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._MONITOR_MODULE"
        )

        with pytest.raises(ExistingMonitorModuleError):
            _set_task_monitor(mock_hub, task_id, sender_id, recipient_id)

        # Check lock was used
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check monitor module is unchanged
        assert monitor._MONITOR_MODULE is mock_monitor

    @pytest.mark.parametrize(
        "initial_state",
        (None, object()),
        ids=("existing_monitor", "no_existing_monitor"),
    )
    def test__unset_task_monitor_gets_existing(
        self,
        initial_state: Optional[object],
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test global monitor module can be unset properly."""
        # Set _MONITOR_MODULE to initial state
        mocker.patch(
            "bitfount.federated.monitoring.monitor._MONITOR_MODULE", initial_state
        )

        _unset_task_monitor()

        # Check lock was used
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check monitor module is removed
        assert monitor._MONITOR_MODULE is None

    def test__get_task_monitor_retrieves_monitor(
        self,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test existing global monitor module can be retrieved."""
        # Set _MONITOR_MODULE to mock
        mock_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._MONITOR_MODULE"
        )

        retrieved_monitor = _get_task_monitor()

        # Check lock was used
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check retrieved monitor is correct monitor
        assert retrieved_monitor is mock_monitor

    def test__get_task_monitor_fails_on_no_existing_monitor(
        self,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test _get_task_monitor fails if no existing global monitor module."""
        # Set _MONITOR_MODULE to None
        mocker.patch("bitfount.federated.monitoring.monitor._MONITOR_MODULE", None)

        # Check error raised and refers to task_monitor_context()
        with pytest.raises(NoMonitorModuleError, match="task_monitor_context"):
            _get_task_monitor()

        # Check lock was used
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

    def test_task_monitor_context(
        self,
        mock_hub: Mock,
        mocker: MockerFixture,
        recipient_id: Optional[str],
        sender_id: str,
        task_id: str,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test that the task_monitor_context correctly creates monitor context."""
        # Set _MONITOR_MODULE to None
        mocker.patch("bitfount.federated.monitoring.monitor._MONITOR_MODULE", None)

        with task_monitor_context(mock_hub, task_id, sender_id, recipient_id):
            # Check monitor exists and is for task
            assert monitor._MONITOR_MODULE is not None
            assert monitor._MONITOR_MODULE.task_id == task_id

            # Check lock was used during creation but is currently unlocked
            wrapped_monitor_lock.acquire.assert_called_once()
            wrapped_monitor_lock.acquire.reset_mock()
            wrapped_monitor_lock.release.assert_called_once()
            wrapped_monitor_lock.release.reset_mock()

        # Check monitor no longer exists
        assert monitor._MONITOR_MODULE is None

        # Check locks were used during deletion but is currently unlocked
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

    def test_task_monitor_context_fails_if_existing_monitor(
        self,
        mock_hub: Mock,
        mocker: MockerFixture,
        recipient_id: Optional[str],
        sender_id: str,
        task_id: str,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test task_monitor_context fails if already in task context."""
        # Set _MONITOR_MODULE to NOT None
        mock_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._MONITOR_MODULE"
        )

        with pytest.raises(ExistingMonitorModuleError):
            with task_monitor_context(mock_hub, task_id, sender_id, recipient_id):
                # We shouldn't reach here as it should fail when entering the
                # task monitor context
                pytest.fail("In task monitor context despite failure")

        # Check monitor is unchanged
        assert monitor._MONITOR_MODULE is mock_monitor

        # Check locks were used during attempted setting but is currently unlocked
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()


@unit_test
class TestMonitorFunctions:
    """Test the functions that interact with the monitor module."""

    def test_task_status_update(
        self,
        mock_monitor_module: Mock,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test task_status_update works correctly."""
        # Patch out _get_task_monitor
        mock_get_task_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._get_task_monitor",
            return_value=mock_monitor_module,
        )

        task_status_update(
            message="Hello World!",
            privacy=MonitorRecordPrivacy.PRIVATE,
            metadata={"meta": "data"},
            progress={"progress": {"value": 1}},
            resource_usage={"resource": 10},
        )

        # Check monitor was retrieved via approved function
        mock_get_task_monitor.assert_called_once()

        # Check locks were used correctly
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check sending to monitor service
        mock_monitor_module.send_to_monitor_service.assert_called_once_with(
            event_type=AdditionalMonitorMessageTypes.TASK_STATUS_UPDATE,
            message="Hello World!",
            privacy=MonitorRecordPrivacy.PRIVATE,
            metadata={"meta": "data"},
            progress={"progress": {"value": 1}},
            resource_usage={"resource": 10},
        )

    def test_task_status_update_handles_no_monitor(
        self,
        caplog: LogCaptureFixture,
        mock_monitor_module: Mock,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test task_status_update handles the case when no monitor exists."""
        # Patch out _get_task_monitor to raise exception because no monitor
        mock_get_task_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._get_task_monitor",
            side_effect=NoMonitorModuleError("TEST ERROR"),
            return_value=mock_monitor_module,
        )

        task_status_update(
            message="Hello World!",
            privacy=MonitorRecordPrivacy.PRIVATE,
            metadata={"meta": "data"},
            progress={"progress": {"value": 1}},
            resource_usage={"resource": 10},
        )

        # Check monitor retrieval was attempted via approved function
        mock_get_task_monitor.assert_called_once()

        # Check locks were used correctly
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check sending to monitor service didn't occur because of error
        mock_monitor_module.send_to_monitor_service.assert_not_called()

        # Check warning was logged
        warning_logs = get_warning_logs(caplog)
        assert "Unable to send task status update: TEST ERROR" in warning_logs

    def test_task_config_update(
        self,
        mock_monitor_module: Mock,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test task_config_update works correctly."""
        # Patch out _get_task_monitor
        mock_get_task_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._get_task_monitor",
            return_value=mock_monitor_module,
        )

        task_config_update(
            task_config={"a": "task", "config": "example"},
            privacy=MonitorRecordPrivacy.PRIVATE,
        )

        # Check monitor was retrieved via approved function
        mock_get_task_monitor.assert_called_once()

        # Check locks were used correctly
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check sending to monitor service
        mock_monitor_module.send_to_monitor_service.assert_called_once_with(
            event_type=AdditionalMonitorMessageTypes.TASK_CONFIG,
            metadata={"a": "task", "config": "example"},
            privacy=MonitorRecordPrivacy.PRIVATE,
        )

    def test_task_config_update_handles_no_monitor(
        self,
        caplog: LogCaptureFixture,
        mock_monitor_module: Mock,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test task_config_update handles the case when no monitor exists."""
        # Patch out _get_task_monitor to raise exception because no monitor
        mock_get_task_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._get_task_monitor",
            side_effect=NoMonitorModuleError("TEST ERROR"),
            return_value=mock_monitor_module,
        )

        task_config_update(
            task_config={"a": "task", "config": "example"},
            privacy=MonitorRecordPrivacy.PRIVATE,
        )

        # Check monitor retrieval was attempted via approved function
        mock_get_task_monitor.assert_called_once()

        # Check locks were used correctly
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check sending to monitor service didn't occur because of error
        mock_monitor_module.send_to_monitor_service.assert_not_called()

        # Check warning was logged
        warning_logs = get_warning_logs(caplog)
        assert "Unable to send task config update: TEST ERROR" in warning_logs


@unit_test
class TestMonitorModule:
    """Tests for the actual _MonitorModule class."""

    @fixture
    def monitor_module(
        self, mock_hub: Mock, recipient_id: str, sender_id: str, task_id: str
    ) -> _MonitorModule:
        """Creates a _MonitorModule instance."""
        return _MonitorModule(
            hub=mock_hub,
            task_id=task_id,
            sender_id=sender_id,
            recipient_id=recipient_id,
        )

    @fixture
    def patch_datetime_now(self, mocker: MockerFixture) -> datetime:
        """Fix a constant datetime instance for tests."""
        current = datetime.now(timezone.utc)
        mock_datetime = mocker.patch("bitfount.federated.monitoring.monitor.datetime")
        mock_datetime.now.return_value = current
        return current

    def test_task_id_property(
        self, monitor_module: _MonitorModule, task_id: str
    ) -> None:
        """Test task_id property returns expected value."""
        assert monitor_module.task_id == task_id

    def test_send_to_monitor_service(
        self,
        mock_hub: Mock,
        monitor_module: _MonitorModule,
        patch_datetime_now: datetime,
        recipient_id: str,
        sender_id: str,
        task_id: str,
    ) -> None:
        """Test send_to_monitor_service works correctly."""
        monitor_module.send_to_monitor_service(
            event_type=AdditionalMonitorMessageTypes.TASK_STATUS_UPDATE,
            privacy=MonitorRecordPrivacy.PRIVATE,
            message="Hello, world!",
            metadata={"meta": "data"},
            progress={"progress": {"value": 1}},
            resource_usage={"resource": 10},
        )

        # Check actual call to hub
        update_json: _MonitorPostJSON = {
            "taskId": task_id,
            "senderId": sender_id,
            "timestamp": patch_datetime_now.isoformat(),
            "privacy": "PRIVATE",
            "type": "TASK_STATUS_UPDATE",
            "message": "Hello, world!",
            "metadata": {"meta": "data"},
            "progress": {"progress": {"value": 1}},
            "resourceUsage": {"resource": 10},
        }
        if recipient_id:
            update_json["recipientId"] = recipient_id
        mock_hub.send_monitor_update.assert_called_once_with(update_json)

    @pytest.mark.parametrize("api_error_cls", (HTTPError, RequestException))
    def test_send_to_monitor_service_handles_api_failure(
        self,
        api_error_cls: Type[Exception],
        caplog: LogCaptureFixture,
        mock_hub: Mock,
        monitor_module: _MonitorModule,
        patch_datetime_now: datetime,
        recipient_id: str,
        sender_id: str,
        task_id: str,
    ) -> None:
        """Test send_to_monitor_service handles error in api call.

        Shouldn't error out, should instead log exception details.
        """
        # Set error condition
        mock_hub.send_monitor_update.side_effect = api_error_cls("ERROR MESSAGE")

        monitor_module.send_to_monitor_service(
            event_type=AdditionalMonitorMessageTypes.TASK_STATUS_UPDATE,
            privacy=MonitorRecordPrivacy.PRIVATE,
            message="Hello, world!",
            metadata={"meta": "data"},
            progress={"progress": {"value": 1}},
            resource_usage={"resource": 10},
        )

        # Check logs
        warning_logs = get_warning_logs(caplog)
        assert (
            "Unable to send monitoring update of type TASK_STATUS_UPDATE;"
            " error was: ERROR MESSAGE" in warning_logs
        )


@pytest.mark.parametrize(
    "message_type",
    # Too many message types in _BitfountMessageType so just choose 5
    argvalues=(
        i
        for i in chain(
            random.sample(list(_BitfountMessageType), 5),
            AdditionalMonitorMessageTypes,
        )
    ),
    ids=lambda x: f"message_type={x.name}",
)
@pytest.mark.parametrize(
    "opt_message",
    argvalues=("Hello, world!", None),
    ids=lambda x: f"message={bool(x)}",
)
@pytest.mark.parametrize(
    "opt_metadata",
    argvalues=({"meta": "data"}, None),
    ids=lambda x: f"metadata={bool(x)}",
)
@pytest.mark.parametrize(
    "opt_progress",
    argvalues=({"progress": {"value": 1}}, None),
    ids=lambda x: f"progress={bool(x)}",
)
@pytest.mark.parametrize(
    "opt_resource_usage",
    argvalues=({"resource": 10}, None),
    ids=lambda x: f"resource={bool(x)}",
)
@pytest.mark.parametrize(
    "privacy",
    argvalues=(i for i in MonitorRecordPrivacy),
    ids=lambda x: f"privacy={x.name}",
)
@unit_test
def test_monitor_post_json_conversion(
    message_type: Union[_BitfountMessageType, AdditionalMonitorMessageTypes],
    opt_message: Optional[str],
    opt_metadata: Optional[_JSONDict],
    opt_progress: Optional[Dict[str, ProgressCounterDict]],
    opt_resource_usage: Optional[Dict[str, float]],
    privacy: MonitorRecordPrivacy,
    recipient_id: Optional[str],
    sender_id: str,
    task_id: str,
) -> None:
    """Test the .json() conversion of _MonitorPostContent."""
    now = datetime.now(timezone.utc)

    mpc = _MonitorPostContent(
        task_id=task_id,
        sender_id=sender_id,
        timestamp=now,
        privacy=privacy,
        type=message_type,
        recipient_id=recipient_id,
        message=opt_message,
        metadata=opt_metadata,
        progress=opt_progress,
        resource_usage=opt_resource_usage,
    )
    mpc_json = mpc.json()

    expected_json: _MonitorPostJSON = {
        "taskId": task_id,
        "senderId": sender_id,
        "timestamp": now.isoformat(),
        "privacy": privacy.value,
        "type": (
            message_type.value
            if isinstance(message_type, AdditionalMonitorMessageTypes)
            else message_type.name
        ),
    }
    if recipient_id:
        expected_json["recipientId"] = recipient_id
    if opt_message:
        expected_json["message"] = opt_message
    if opt_metadata:
        expected_json["metadata"] = opt_metadata
    if opt_progress:
        expected_json["progress"] = opt_progress
    if opt_resource_usage:
        expected_json["resourceUsage"] = opt_resource_usage

    assert mpc_json == expected_json


@unit_test
class TestMonitorDecorators:
    """Tests for the monitor decorator functions."""

    @task_status("status update test")
    def _simple_func(self, y: int) -> int:
        return y + 1

    def test_task_status_decorator(
        self,
        mock_monitor_module: Mock,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test the task_status decorator works."""
        # Patch out _get_task_monitor
        mock_get_task_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._get_task_monitor",
            return_value=mock_monitor_module,
        )

        x = self._simple_func(1)

        # Check monitor was retrieved via approved function
        mock_get_task_monitor.assert_called_once()

        # Check locks were used correctly
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check sending to monitor service
        mock_monitor_module.send_to_monitor_service.assert_called_once_with(
            event_type=AdditionalMonitorMessageTypes.TASK_STATUS_UPDATE,
            message="status update test",
            privacy=MonitorRecordPrivacy.ALL_PARTICIPANTS,
            metadata=None,
            progress=None,
            resource_usage=None,
        )

        # Check function return unaffected
        assert x == 2

    def test_task_status_decorator_handles_error(
        self,
        caplog: LogCaptureFixture,
        mock_monitor_module: Mock,
        mocker: MockerFixture,
        wrapped_monitor_lock: Mock,
    ) -> None:
        """Test the task_status decorator works."""
        # Patch out _get_task_monitor
        mock_get_task_monitor = mocker.patch(
            "bitfount.federated.monitoring.monitor._get_task_monitor",
            side_effect=NoMonitorModuleError("TEST ERROR"),
            return_value=mock_monitor_module,
        )

        x = self._simple_func(1)

        # Check monitor was retrieved via approved function
        mock_get_task_monitor.assert_called_once()

        # Check locks were used correctly
        wrapped_monitor_lock.acquire.assert_called_once()
        wrapped_monitor_lock.release.assert_called_once()

        # Check sending to monitor service didn't occur because of error
        mock_monitor_module.send_to_monitor_service.assert_not_called()

        # Check warning was logged
        warning_logs = get_warning_logs(caplog)
        assert "Unable to send task status update: TEST ERROR" in warning_logs

        # Check function return unaffected
        assert x == 2



"""Defines abstract models, mixins, and other common backend-agnostic classes.

Implementations of these abstract models should be located in `bitfount.models.models`
or in the `models` subpackage of a backend.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, auto
import inspect
import logging
import os
from pathlib import Path
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import desert
from marshmallow import fields
import numpy as np

from bitfount.config import BITFOUNT_LOGS_DIR
from bitfount.data.databunch import BitfountDataBunch
from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.datastructure import registry as datastructure_registry
from bitfount.data.schema import BitfountSchema, TableSchema
from bitfount.types import (
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    T_OPTIONAL_NESTED,
    DistributedModelProtocol,
    _BaseSerializableObjectMixIn,
    _StrAnyDict,
)
from bitfount.utils import _add_this_to_list, delegates, seed_all

if TYPE_CHECKING:
    from bitfount.metrics import Metric


# Valid pooling functions for Convolutional Neural Networks.
POOLING_FUNCTIONS = ["max", "avg"]


logger = logging.getLogger(__name__)

# Main Model registry that is used to collect all model classes.
# Used for serialization and deserialization of model classes.
# Includes both abstract and non-abstract classes.
MAIN_MODEL_REGISTRY: _StrAnyDict = {}

# MYPY ISSUES
# Due to the use of desert.field to create non-standard fields in these dataclasses,
# mypy complains about incompatible type assignment. This is a known bug in desert
# (https://github.com/python-desert/desert/issues/101) and so we can ignore it
# until the issue is resolved.


@dataclass
class LoggerConfig:
    """Configuration for the logger.

    The configured logger will log training events, metrics, model checkpoints, etc. to
    your chosen platform. If no logger configuration is provided, the default logger is
    a Tensorboard logger.

    Args:
        name: The name of the logger. Should be one of the loggers supported by the
            chosen backend
        save_dir: The directory to save the logs. Defaults to `BITFOUNT_LOGS_DIR`
        params: A dictionary of keyword arguments to pass to the logger. Defaults to an
            empty dictionary
    """

    #: same as argument
    name: str
    #: same as argument
    save_dir: Optional[Path] = desert.field(
        fields.Function(
            deserialize=lambda path: path if path is None else Path(path).expanduser()
        ),
        default=BITFOUNT_LOGS_DIR,
    )  # type: ignore[assignment] # Reason: see top of file
    #: same as argument
    params: Optional[_StrAnyDict] = desert.field(
        fields.Dict(keys=fields.Str), default_factory=dict
    )  # type: ignore[assignment] # Reason: see top of file


class ModelContext(Enum):
    """Describes the context (modeller or worker) in which the model is running.

    This is used for models where the model differs depending on if it is on the
    modeller-side or worker-side of the federated process.
    """

    MODELLER = auto()
    WORKER = auto()


ModelType = TypeVar("ModelType", bound="_BaseModel")
_model_structure_registry: Dict[str, Type[_BaseModelStructureMixIn]] = {}
model_structure_registry: Mapping[
    str, Type[_BaseModelStructureMixIn]
] = MappingProxyType(_model_structure_registry)


@dataclass
class _BaseModelStructureMixIn:
    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
            _model_structure_registry[cls.__name__] = cls


class _BaseModelRegistryMixIn:
    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if cls.__name__ not in MAIN_MODEL_REGISTRY:
            MAIN_MODEL_REGISTRY[cls.__name__] = cls


class _BaseModel(
    _BaseModelRegistryMixIn, _BaseSerializableObjectMixIn, ABC, Generic[ModelType]
):
    """Abstract Base Model from which all other models must inherit.

    This class is designed to be at the very bottom of the inheritance hierarchy.
    The only reason it has a `super().__init__()` call is to call the parent classes of
    other classes defined in other libraries. It also takes kwargs so that we do not
    throw an error if there are unexpected keyword arguments. These unexpected keyword
    arguments will end up in this constructor where they will simply be ignored.

    Args:
        datastructure: `DataStructure` to be passed to the model when initialised
        schema: The `BitfountSchema` object associated with the datasource
            on which the model will be trained on.
        seed: Random number seed. Used for setting random seed for all libraries.
            Defaults to None.
        param_clipping: Arguments for clipping for BatchNorm parameters.
            Used for federated models with secure aggregation.
            It should contain the SecureShare variables and the
            number of workers in a dictionary,
            e.g. {"prime_q":13, "precision": 10**3,"num_workers":2}.
            Defaults to None.

    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "seed": fields.Integer(allow_none=True),
        "schema": fields.Nested(BitfountSchema._Schema),
        "param_clipping": fields.Dict(
            keys=fields.String(), values=fields.Integer(), allow_none=True
        ),
    }
    # TODO: [BIT-1623] BitfountSchema here temporarily as a nested field,
    #  should be removed following the RFC
    nested_fields: ClassVar[T_NESTED_FIELDS] = {"datastructure": datastructure_registry}

    def __init__(
        self,
        datastructure: DataStructure,
        schema: BitfountSchema,
        seed: Optional[int] = None,
        param_clipping: Optional[Dict[str, int]] = None,
        **kwargs: Any,
    ):
        self.class_name = f"bitfount.{type(self).__name__}"
        self._context: Optional[ModelContext] = None
        self.metrics: Optional[MutableMapping[str, Metric]] = None
        self._model: Optional[ModelType] = None
        self._initialised: bool = False
        self.seed = seed
        self.param_clipping = param_clipping
        seed_all(self.seed)
        self.datastructure = datastructure
        self.databunch: BitfountDataBunch
        self.schema = schema
        self._objective: str

        # Placeholders for dataloaders
        self.train_dl: BitfountDataLoader
        self.validation_dl: Optional[BitfountDataLoader] = None
        self.test_dl: Optional[BitfountDataLoader] = None

        for unexpected_kwarg in kwargs:
            logger.warning(f"Ignoring unexpected keyword argument {unexpected_kwarg}")

        super().__init__()

    def _set_dataloaders(
        self,
        batch_size: Optional[int] = None,
    ) -> None:
        """Sets train, validation and test dataloaders.

        Args:
            batch_size: The batch size to use for the dataloaders. Defaults to None.
        """
        if self.databunch is None:
            raise ValueError(
                "_set_dataloaders() requires the databunch to be set "
                "before being called."
            )
        self.train_dl = self.databunch.get_train_dataloader(batch_size)
        self.validation_dl = self.databunch.get_validation_dataloader(batch_size)
        self.test_dl = self.databunch.get_test_dataloader(batch_size)

    def _add_datasource_to_schema(
        self,
        datasource: BaseSource,
    ) -> None:
        """Adds the datasource to the schema.

        Should be only used for local training.
        """
        # self.datastructure.apply_transformations(datasource)
        # TODO: [BIT-1167] Once we add the transformations we can force the target
        #  to be categorical in the transformations as this can be specified directly
        if self.datastructure.image_cols:
            datasource.load_data()
            datasource_columns = datasource.get_dtypes()
            for image in self.datastructure.image_cols:
                if image not in datasource_columns:
                    raise ValueError(f"Could not find {image} in dataset columns")

        # If the model is distributed, then we need to get the table name that
        # the Modeller has specified for this particular Pod.
        pod_identifier: Optional[str] = None
        if hasattr(self, "pod_identifier"):
            # Assuring mypy that the attribute exists
            assert isinstance(self, DistributedModelProtocol)  # nosec[assert_used]
            pod_identifier = self.pod_identifier
        table_name = self.datastructure.get_table_name(pod_identifier)

        # Add the datasource features to the datastructure.
        # This intentionally does not support multi-table datasources because a
        # multi-table datasource should have been reduced to a single table by this
        # point
        self.schema.add_datasource_tables(
            datasource,
            ignore_cols={table_name: self.datastructure.ignore_cols},
            force_stypes={table_name: self.datastructure._force_stype},
            table_name=table_name,
        )
        self.databunch = BitfountDataBunch(
            data_structure=self.datastructure,
            schema=self.schema.get_table_schema(table_name),
            datasource=datasource,
        )

    def initialise_model(
        self,
        data: Optional[BaseSource] = None,
        context: Optional[ModelContext] = None,
    ) -> None:
        """Can be implemented to initialise model if necessary.

        This is automatically called by the `fit()` method if necessary.

        Args:
            data: The data used for model training.
            context: Indicates if the model is running as a modeller or worker. If None,
                there is no difference between modeller and worker. Defaults to None.
        """
        self._initialised = True
        if context is None and data is not None:
            self._add_datasource_to_schema(datasource=data)

    @property
    def initialised(self) -> bool:
        """Returns True if the model has been initialised, otherwise False.

        I.e. whether the `initialise_model` method has been called.
        """
        return self._initialised

    @abstractmethod
    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Implement this method to serialise a model."""
        raise NotImplementedError

    @abstractmethod
    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Implement this method to deserialise a model."""
        raise NotImplementedError

    @abstractmethod
    def evaluate(
        self,
        test_dl: Optional[BitfountDataLoader] = None,
        pod_identifiers: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Dict[str, float]]:
        """Implement this method to perform inference on the test set.

        Args:
            test_dl: Optional `BitfountDataLoader` object containing test data. If this
                is not provided, the test set from the `BaseSource` used to train the
                model should be used if present.
            pod_identifiers: A list of pod identifiers to use for remote evaluation.

        Returns:
            A tuple of numpy arrays containing the predicted and actual values.
        """
        raise NotImplementedError

    def predict(self, data: BaseSource, **kwargs: Any) -> np.ndarray:
        """This method runs inference on the test data, returns predictions.

        Args:
            data: `BaseSource` object containing the data to run prediction on.
                Predictions will be generated for the test subset (as defined
                by the `DataSetSplitter`).

        Returns:
            A numpy array containing the prediction values.
        """
        raise NotImplementedError

    @abstractmethod
    def fit(
        self,
        data: Optional[BaseSource] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        pod_identifiers: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Optional[Dict[str, str]]:
        """Must be implemented to fit the model.

        Must call `initialise_model()` within the method if the model needs to be
        initialised.

        Args:
            data: The data used for local model training.
            metrics: A dictionary of metrics to use for validation.
            pod_identifiers: A list of pod identifiers to use for federated training.

        """
        raise NotImplementedError


class ClassifierMixIn(_BaseModelRegistryMixIn, _BaseSerializableObjectMixIn):
    """MixIn for classification problems.

    Classification models must have this class in their inheritance hierarchy.

    Args:
        multilabel: Whether the problem is a multi-label problem. i.e. each datapoint
            belongs to multiple classes
        param_clipping: Arguments for clipping for BatchNorm parameters.
            Used for federated models with secure aggregation.
            It should contain the SecureShare variables and the
            number of workers in a dictionary,
            e.g. {"prime_q":13, "precision": 10**3,"num_workers":2}


    Attributes:
        multilabel: Whether the problem is a multi-label problem
        n_classes: Number of classes in the problem
    """

    #: set in _BaseModel
    datastructure: DataStructure
    #: set in _BaseModel
    schema: BitfountSchema
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "multilabel": fields.Bool(),
        "param_clipping": fields.Dict(
            keys=fields.String(), values=fields.Integer(), allow_none=True
        ),
    }
    nested_fields: ClassVar[T_NESTED_FIELDS] = {"datastructure": datastructure_registry}

    def __init__(
        self,
        multilabel: bool = False,
        param_clipping: Optional[Dict[str, int]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.multilabel = multilabel
        self.param_clipping = param_clipping
        self.n_classes: int
        if "n_classes" in kwargs:
            self.n_classes = int(kwargs["n_classes"])
        self._objective = "classification"

    def set_number_of_classes(self, schema: TableSchema) -> None:
        """Sets the target number of classes for the classifier.

        If the data is a multi-label problem, the number of classes is set to the number
        of target columns as specified in the `DataStructure`. Otherwise, the number of
        classes is set to the number of unique values in the target column as specified
        in the `BitfountSchema`. The value is stored in the `n_classes` attribute.
        """
        if self.datastructure.target is None and hasattr(self, "n_classes"):
            logger.warning(
                "No target specified in data. Using explicitly provided n_classes."
                "Note that only inference results will be valid, not training"
                "or evaluation for this model and dataset."
            )
        elif self.datastructure.target is not None:
            # If the model is distributed, then we need to get the table name that
            # the Modeller has specified for this particular Pod.
            self.n_classes = (
                len(self.datastructure.target)
                if self.multilabel
                else schema.get_categorical_feature_size(self.datastructure.target)
            )
        else:
            raise ValueError(
                "No target specified in data, and number of classes not specified "
                "explicitly. Not able to determine dimensions of head of model."
            )

    def _add_datasource_to_schema(
        self,
        datasource: BaseSource,
    ) -> None:
        """Adds the datasource to the schema.

        Used for local training.
        """
        # Add the datasource features to the datastructure.
        force_categorical: List[str] = []
        force_categorical = _add_this_to_list(
            self.datastructure.target, force_categorical
        )
        force_categorical = _add_this_to_list(
            self.datastructure.multihead_col, force_categorical
        )
        if self.datastructure.image_cols:
            datasource.load_data()
            datasource_columns = datasource.get_dtypes()
            for image in self.datastructure.image_cols:
                if image not in datasource_columns:
                    raise ValueError(f"Could not find {image} in dataset columns")

        if "categorical" not in self.datastructure._force_stype:
            self.datastructure._force_stype["categorical"] = force_categorical
        else:
            _add_this_to_list(
                force_categorical, self.datastructure._force_stype["categorical"]
            )

        # This intentionally does not support multi-table datasources because a
        # multi-table datasource should have been reduced to a single table by this
        # point
        table_name = self.datastructure.get_table_name()
        self.schema.add_datasource_tables(
            datasource,
            ignore_cols={table_name: self.datastructure.ignore_cols},
            force_stypes={table_name: self.datastructure._force_stype},
            table_name=table_name,
        )
        self.databunch = BitfountDataBunch(
            data_structure=self.datastructure,
            schema=self.schema.get_table_schema(table_name),
            datasource=datasource,
        )


class RegressorMixIn(_BaseModelRegistryMixIn, _BaseSerializableObjectMixIn):
    """MixIn for regression problems.

    Currently, just used for tagging purposes.
    """

    pass


@dataclass
class NeuralNetworkModelStructure(
    ABC, _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
):
    """Dataclass defining the structure of a neural network model.

    Args:
        layers: List of hidden layer sizes i.e. not including the input/output layers.
            The layer type depends on the specific model structure. Defaults to 2 layers
            with 1000 and 500 nodes respectively.
        dropout_probs: List of dropout probabilities for each layer. Must be the same
            length as `layers`. Defaults to 0.01 and 0.1 respectively.
        mish_activation_function: Whether to use Mish activation function. If False,
            ReLU is used instead. Defaults to True.

    Attributes:
        num_heads: Number of heads for multihead models. I.e. the number of output
            layers.

    Raises:
        ValueError: If `layers` and `dropout_probs` are not the same length.
    """

    #: same as argument
    layers: Optional[List[int]] = None
    #: same as argument
    dropout_probs: Optional[List[float]] = None
    #: same as argument
    mish_activation_function: bool = True

    fields_dict = {
        "layers": fields.List(fields.Int()),
        "dropout_probs": fields.List(fields.Float()),
        "mish_activation_function": fields.Bool(),
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]

    def __post_init__(self) -> None:
        self.class_name = type(self).__name__
        if self.layers is None:
            self.layers = [1000, 500]
        if self.dropout_probs is None:
            self.dropout_probs = [0.01, 0.1]
        if len(self.layers) != len(self.dropout_probs):
            raise ValueError(
                "Number of neural layers must equal number of dropout layers."
            )
        self.num_heads: int

    def set_num_heads(self, datastructure: DataStructure) -> None:
        """Sets the `num_heads` attribute.

        This is taken from the `multihead_size` attribute of the `datastructure`
        if present. Otherwise, it is set to 1.

        Args:
            datastructure: The data structure the model has been designed for.

        Raises:
            ValueError: If `datastructure.multihead_size` is not a positive integer.

        """
        self.num_heads = (
            datastructure.multihead_size
            if datastructure.multihead_size is not None
            else 1
        )


@delegates()
@dataclass
class FeedForwardModelStructure(
    NeuralNetworkModelStructure, _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
):
    """Dataclass defining the structure of a feedforward neural network model.

    This model structure only defines linear and dropout layers.

    Args:
        embedding_dropout: Dropout probability for embedding layer. Defaults to 0.04.
    """

    #: same as argument
    embedding_dropout: float = 0.04
    fields_dict = {
        "embedding_dropout": fields.Float(default=0.04)
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]


@delegates()
@dataclass
class CNNModelStructure(
    NeuralNetworkModelStructure, _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
):
    """Dataclass defining the structure of a convolutional neural network model.

    This model structure has multiple convolutional layers followed by multiple
    linear and dropout layers.

    Args:
        kernel_size: Kernel size for convolutional layers. Defaults to 5.
        padding: Padding for convolutional layers. Defaults to 2.
        stride: Stride for convolutional layers. Defaults to 1.
        pooling_function: Pooling function for convolutional layers. Options are "max"
            or "avg". Defaults to "max".
        ff_layers: List of linear hidden layer sizes following the convolutional layers.
            Defaults to [1000, 500].
        ff_dropout_probs: List of dropout probabilities for each linear layer. Must be
            the same length as `ff_layers`. Defaults to 0.01 and 0.1 respectively.

    Raises:
        ValueError: If `ff_layers` and `ff_dropout_probs` are not the same length.
        ValueError: If `pooling_function` is not "max" or "avg".
    """

    #: same as argument
    kernel_size: int = 5
    #: same as argument
    padding: int = 2
    #: same as argument
    stride: int = 1
    #: same as argument
    pooling_function: str = "max"
    #: same as argument
    ff_layers: List[int] = field(default_factory=list)
    #: same as argument
    ff_dropout_probs: List[float] = field(default_factory=list)
    fields_dict = {
        "kernel_size": fields.Int(),
        "padding": fields.Int(),
        "stride": fields.Int(),
        "pooling_function": fields.Str(),
        "layers": fields.List(fields.Int()),
        "ff_layers": fields.List(fields.Int(), allow_none=True, default=[1000, 500]),
        "ff_dropout_probs": fields.List(
            fields.Float(), allow_none=True, default=[0.01, 0.1]
        ),
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]

    def __post_init__(self) -> None:
        if not isinstance(self.layers, list):
            self.layers = [16, 32]
        super().__post_init__()
        if self.pooling_function not in POOLING_FUNCTIONS:
            raise ValueError(
                f"Pooling function can only be one of: {POOLING_FUNCTIONS}"
            )
        self.ff_layers = [1000, 500] if len(self.ff_layers) == 0 else self.ff_layers
        self.ff_dropout_probs = (
            [0.01, 0.1] if len(self.ff_dropout_probs) == 0 else self.ff_dropout_probs
        )
        if len(self.ff_layers) != len(self.ff_dropout_probs):
            raise ValueError(
                "Number of neural layers must equal number of dropout layers."
            )


@dataclass
class NeuralNetworkPredefinedModel(
    _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
):
    """This class encompasses what is required to use a predefined model e.g. ResNet.

    The currently supported models are:
        - AlexNet
        - DenseNet{121,161,169,201}
        - ResNet{18,34,50,101,152}
        - SqueezeNet{1_0,1_1}
        - VGG{11,11_bn,13,13_bn,16,16_bn,19,19_bn}
        - TabNet

    Args:
        name: name of the model
        pretrained: flag to denote whether to download the pretrained parameters
        **kwargs: additional arguments to pass to the model constructor
    """

    #: same as argument
    name: str
    #: same as argument
    pretrained: bool = True
    #: same as argument
    kwargs: Optional[Dict[str, bool]] = None
    fields_dict = {
        "name": fields.String(),
        "pretrained": fields.Bool(),
        "kwargs": fields.Dict(keys=fields.Str(), values=fields.Bool(), allow_none=True),
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]

    def __post_init__(self) -> None:
        self.class_name = type(self).__name__


_optimizer_registry: Dict[str, Type[_BaseOptimizer]] = {}
optimizer_registry: Mapping[str, Type[_BaseOptimizer]] = MappingProxyType(
    _optimizer_registry
)


@dataclass
class _BaseOptimizer:
    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
            _optimizer_registry[cls.__name__] = cls


@dataclass
class Optimizer(_BaseOptimizer, _BaseSerializableObjectMixIn):
    """Class for specifying the optimizer for a neural network.

    The options for the optimizer will depend on the backend being used.

    Args:
        name: name of the optimizer
        params: dictionary of keyword arguments for the optimizer constructor
    """

    #: same as argument
    name: str
    #: same as argument
    params: _StrAnyDict = field(default_factory=dict)
    fields_dict = {
        "name": fields.String(),
        "params": fields.Dict(),
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]

    def __post_init__(self) -> None:
        self.class_name = type(self).__name__


_scheduler_registry: Dict[str, Type[_BaseScheduler]] = {}
scheduler_registry: Mapping[str, Type[_BaseScheduler]] = MappingProxyType(
    _scheduler_registry
)


@dataclass
class _BaseScheduler:
    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
            _scheduler_registry[cls.__name__] = cls


@dataclass
class Scheduler(_BaseScheduler, _BaseSerializableObjectMixIn):
    """Class for specifying the scheduler for a neural network.

    The options for the scheduler will depend on the backend being used.

    Args:
        name: name of the scheduler
        params: dictionary of keyword arguments for the scheduler constructor
    """

    #: same as argument
    name: str
    #: same as argument
    params: _StrAnyDict = field(default_factory=dict)
    fields_dict = {
        "name": fields.String(),
        "params": fields.Dict(),
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]

    def __post_init__(self) -> None:
        self.class_name = type(self).__name__


_early_stopping_registry: Dict[str, Type[_BaseEarlyStopping]] = {}
early_stopping_registry: Mapping[str, Type[_BaseEarlyStopping]] = MappingProxyType(
    _early_stopping_registry
)


@dataclass
class _BaseEarlyStopping:
    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
            _early_stopping_registry[cls.__name__] = cls


@dataclass
class EarlyStopping(_BaseEarlyStopping, _BaseSerializableObjectMixIn):
    """Class for specifying early stopping parameters.

    Args:
        params: dictionary of keyword arguments for the early stopping constructor. The
            parameters will depend on the backend being used.
    """

    #: same as argument
    params: _StrAnyDict = field(default_factory=dict)
    fields_dict = {
        "params": fields.Dict(),
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]

    def __post_init__(self) -> None:
        self.class_name = type(self).__name__


class NeuralNetworkMixIn(_BaseSerializableObjectMixIn):
    """Specifies model structure and hyperparameters for neural network models.

    All neural network models must inherit from this class.

    :::caution

    `custom_loss_func` cannot be serialized and therefore cannot be used in Federated
    Learning.

    :::

    Args:
        model_structure: The structure of the model.
        batch_size: The number of data points in each batch. Defaults to 512.
        epochs: The number of epochs to train for. If `steps` is provided, `epochs`
            cannot be provided. Defaults to None.
        steps: The number of steps to train for. If `epochs` is provided, `steps`
            cannot be provided. Defaults to None.
        optimizer: The optimizer to use. Defaults to "AdamW" with a learning rate of
            0.01.
        scheduler: The scheduler to use. Defaults to None.
        custom_loss_func: A custom loss function to use. Defaults to None.
        early_stopping: Early stopping parameters. Defaults to None.

    Attributes:
        model_structure: The structure of the model.
        batch_size: The number of data points in each batch.
        epochs: The number of epochs to train for.
        steps: The number of steps to train for.
        optimizer: The optimizer to use.
        scheduler: The scheduler to use.
        loss_func: A custom loss function to use.
        early_stopping: Early stopping parameters.

    Raises:
        ValueError: If both `epochs` and `steps` are provided.
    """

    #: set in _BaseModel
    datastructure: DataStructure
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "epochs": fields.Integer(allow_none=True),
        "steps": fields.Integer(allow_none=True),
        "batch_size": fields.Integer(allow_none=True),
        "custom_loss_func": fields.Raw(allow_none=True),
    }
    nested_fields: ClassVar[T_NESTED_FIELDS] = {
        "datastructure": datastructure_registry,
        "model_structure": model_structure_registry,
        "optimizer": optimizer_registry,
        "scheduler": scheduler_registry,
        "early_stopping": early_stopping_registry,
    }
    optional_nested: ClassVar[T_OPTIONAL_NESTED] = ["scheduler", "early_stopping"]

    def __init__(
        self,
        model_structure: Union[
            NeuralNetworkModelStructure, NeuralNetworkPredefinedModel
        ],
        batch_size: int = 512,
        epochs: Optional[int] = None,
        steps: Optional[int] = None,
        optimizer: Optional[Optimizer] = None,
        scheduler: Optional[Scheduler] = None,
        custom_loss_func: Optional[Callable] = None,
        early_stopping: Optional[EarlyStopping] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        if (steps is None and epochs is None) or (
            isinstance(steps, int) and isinstance(epochs, int)
        ):
            raise ValueError("You must specify one (and only one) of steps or epochs.")

        if optimizer is None:
            logger.info("No optimizer provided, using AdamW with learning rate 1e-2.")
            self.optimizer = Optimizer("AdamW", {"lr": 0.01})
        elif isinstance(optimizer, Optimizer):
            self.optimizer = optimizer

        self.model_structure = model_structure
        self.epochs = epochs
        self.steps = steps
        self.batch_size = batch_size
        self.scheduler = scheduler
        self._opt_func: Optional[Callable] = None  # To be set by initialise_model()
        self._scheduler_func: Optional[
            Callable
        ] = None  # To be set by initialise_model()
        self.loss_func: Optional[Callable] = custom_loss_func  # Only works locally
        self.early_stopping = early_stopping

        if isinstance(self.model_structure, NeuralNetworkModelStructure):
            self.model_structure.set_num_heads(self.datastructure)

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if cls.__name__ not in MAIN_MODEL_REGISTRY:
            MAIN_MODEL_REGISTRY[cls.__name__] = cls

    @staticmethod
    @abstractmethod
    def _get_optimizer(optimizer: Optimizer) -> Callable:
        """Returns appropriate optimizer class."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _get_scheduler(scheduler: Scheduler) -> Callable:
        """Returns appropriate scheduler class."""
        raise NotImplementedError



"""Modules for data sources.

Datasource plugins can also be imported from this package.
"""
import logging as _logging
import pkgutil as _pkgutil

from bitfount.config import BITFOUNT_PLUGIN_PATH as _BITFOUNT_PLUGIN_PATH
from bitfount.utils import _get_module_from_file

_logger = _logging.getLogger(__name__)

# Create `datasources` plugin subdir if it doesn't exist
_datasource_plugin_path = _BITFOUNT_PLUGIN_PATH / "datasources"
_datasource_plugin_path.mkdir(parents=True, exist_ok=True)

# Add datasource plugin modules to the `datasources` namespace alongside the existing
# built-in datasource modules. This is not essential, but it allows users to import
# the entire plugin module as opposed to just the Datasource class which is what is done
# in the `bitfount.data` __init__ module.
for _module_info in _pkgutil.walk_packages(
    [str(_datasource_plugin_path)],
):
    try:
        _module = _get_module_from_file(
            _datasource_plugin_path / f"{_module_info.name}.py"
        )
        globals().update({_module.__name__: _module})
        _logger.info(f"Loaded datasource plugin: {_module_info.name}")
    except ImportError:
        pass



"""Classes concerning data loading and dataloaders."""
from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any, Iterator, List, Optional, Tuple, Union

import pandas as pd

if TYPE_CHECKING:
    from bitfount.data.datasets import _BaseBitfountDataset
    from bitfount.data.types import _SingleOrMulti


class BitfountDataLoader:
    """An backend-agnostic data loader.

    Args:
        dataset: The dataset for the dataloader.
        batch_size: The batch size for the dataloader.
            Defaults to None.
    """

    def __init__(self, dataset: _BaseBitfountDataset, batch_size: Optional[int] = None):
        self.dataset = dataset
        self.batch_size = batch_size

    def __len__(self) -> int:
        """Number of batches or number of elements if batch size is None."""
        if not self.batch_size:
            return len(self.dataset)
        return math.ceil(len(self.dataset) / self.batch_size)

    def __iter__(self) -> Iterator[List[_SingleOrMulti[Any]]]:
        """This should be implemented to allow batch by batch loading.

        Currently there are no backend-agnostic models that can operate on iterable
        datasets so it has not been implemented.

        Returns:
            An iterator over batches of x and y numpy arrays.
        """
        raise NotImplementedError

    def get_x_dataframe(
        self,
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        """Gets the x-dataframe of the data i.e. features.

        For models incompatible with the __iter__ approach.
        """
        tabular, image, support_cols = self.dataset.x_var
        if tabular.size and image.size:
            tab_cols = [
                col
                for col in self.dataset.x_columns
                if col not in self.dataset.image_columns
            ]
            tab_df = pd.DataFrame(data=tabular, columns=tab_cols)
            tab_df[self.dataset.embedded_col_names] = tab_df[
                self.dataset.embedded_col_names
            ].astype("int64")
            img_df = pd.DataFrame(data=image, columns=self.dataset.image_columns)
            return tab_df, img_df
        elif image.size:
            img_df = pd.DataFrame(data=image, columns=self.dataset.image_columns)
            return img_df
        elif tabular.size:
            columns = self.dataset.x_columns
            tab_df = pd.DataFrame(data=tabular, columns=columns)
            tab_df[self.dataset.embedded_col_names] = tab_df[
                self.dataset.embedded_col_names
            ].astype("int64")
            return tab_df
        else:
            raise ValueError("No tabular or image data to train with.")

    def get_y_dataframe(self) -> pd.DataFrame:
        """Gets the y-dataframe of the data i.e. target.

        For models incompatible with the __iter__ approach.
        """
        columns = self.dataset.y_columns
        data = self.dataset.y_var
        dataframe = pd.DataFrame(data=data, columns=columns)
        return dataframe



"""PyTorch implementations for Bitfount Dataset classes."""
from typing import Sequence, Union

import torch
from torch.utils.data import Dataset as PTDataset
from torch.utils.data import IterableDataset as PTIterableDataset

from bitfount.backends.pytorch.data.utils import _index_tensor_handler
from bitfount.data.datasets import _BitfountDataset, _IterableBitfountDataset
from bitfount.data.types import _DataEntry


class _PyTorchDataset(_BitfountDataset, PTDataset):
    """See base class."""

    def __getitem__(self, idx: Union[int, Sequence[int], torch.Tensor]) -> _DataEntry:
        idx = _index_tensor_handler(idx)
        return self._getitem(idx)


class _PyTorchIterableDataset(_IterableBitfountDataset, PTIterableDataset):
    """See base class."""

    pass



"""Module containing ExcelSource class.

ExcelSource class handles loading of Excel data.
"""
import logging
import os
from typing import Any, Dict, Iterable, List, Optional, Union, cast

import pandas as pd
from pydantic import AnyUrl

from bitfount.data.datasources.base_source import MultiTableSource
from bitfount.data.exceptions import ExcelSourceError
from bitfount.data.types import _SingleOrMulti
from bitfount.types import _Dtypes
from bitfount.utils import delegates

logger = logging.getLogger(__name__)


@delegates()
class ExcelSource(MultiTableSource):
    """Data source for loading excel files.

    :::info

    You must install a backend library to read excel files to use this data source.
    Currently supported engines are “xlrd”, “openpyxl”, “odf” and “pyxlsb”.

    :::

    :::info

    By default, the first row is used as the column names unless `column_names` or the
    `header` keyword argument is provided.

    :::

    Args:
        path: The path or URL to the excel file.
        sheet_name: The name(s) of the sheet(s) to load. If not provided, the
            all sheets will be loaded.
        column_names: The names of the columns if not using the first row of the sheet.
            Can only be used for single sheet excel files.
        dtype: The dtypes of the columns.
        **read_excel_kwargs: Additional arguments to be passed to `pandas.read_excel`.

    Raises:
        TypeError: If the path does not have the correct extension denoting an excel
            file.
        ValueError: If multiple sheet names are provided and column names are also
            provided.
        ValueError: If sheets are referenced which do not exist in the excel file.

    """

    def __init__(
        self,
        path: Union[os.PathLike, AnyUrl, str],
        sheet_name: Optional[_SingleOrMulti[str]] = None,
        column_names: Optional[List[str]] = None,
        dtype: Optional[_Dtypes] = None,
        read_excel_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        if not str(path).endswith(
            (".xls", ".xlsx", "xlsm", "xlsb", "odf", "ods", "odt")
        ):
            raise TypeError("Please provide a Path or URL to an Excel file.")
        self.file = pd.ExcelFile(str(path))
        self.sheet_name = sheet_name

        if not set(self.table_names).issubset(set(self.file.sheet_names)):
            raise ValueError(
                f"Sheet(s) "
                f"{', '.join([i for i in self.table_names if i not in self.file.sheet_names])}"  # noqa: B950
                f" were not found in the Excel file."
            )

        if not read_excel_kwargs:
            read_excel_kwargs = {}

        if self.multi_table and column_names:
            raise ValueError(
                "Column names can only be provided "
                "if a single sheet name is provided."
            )

        if column_names:
            read_excel_kwargs["names"] = column_names
            read_excel_kwargs["header"] = None

        if dtype:
            read_excel_kwargs["dtype"] = dtype

        self.read_excel_kwargs = read_excel_kwargs
        self.datastructure_table_name: Optional[str] = None

    @property
    def table_names(self) -> List[str]:
        """Excel sheet names in datasource."""
        if self.sheet_name is not None:
            if isinstance(self.sheet_name, list):
                return [i for i in self.sheet_name]

            elif isinstance(self.sheet_name, str):
                return [self.sheet_name]

        return cast(List[str], self.file.sheet_names)

    def _validate_table_name(self, table_name: str) -> None:
        """Validate the table name exists in the excel workbook.

        Args:
            table_name: The name of the table.

        Raises:
            ValueError: If the table name is not found in the data.
        """
        # This function should always be called with a table name.
        if table_name not in self.table_names:
            raise ValueError(
                f"Table name {table_name} not found in the data. "
                f"Available tables: {', '.join(self.table_names)}"
            )

    def get_data(
        self,
        table_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Optional[pd.DataFrame]:
        """Loads and returns data from Excel dataset.

        Args:
            table_name: Table name for multi table data sources. This
                comes from the DataStructure and is ignored if sql_query has been
                provided.

        Returns:
            A DataFrame-type object which contains the data.

        Raises:
            ValueError: If the table name provided does not exist.
        """
        df: Optional[pd.DataFrame] = None
        if not self.multi_table:
            # If `sheet_name` is not a string, `pd.read_excel` returns a dictionary
            # of DataFrames instead of a single DataFrame so we ensure it is a string
            if isinstance(self.sheet_name, str):
                table_name = self.sheet_name
            elif isinstance(self.sheet_name, list):
                table_name = self.sheet_name[0]
            else:
                table_name = self.table_names[0]

            df = pd.read_excel(
                self.file, sheet_name=table_name, **self.read_excel_kwargs
            )

        elif table_name:
            self._validate_table_name(table_name)
            df = pd.read_excel(
                self.file, sheet_name=table_name, **self.read_excel_kwargs
            )
        # If no table name specified we don't load anything into the dataframe.
        return df

    def get_values(
        self,
        col_names: List[str],
        table_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Iterable[Any]]:
        """Get distinct values from columns in Excel dataset.

        Args:
            col_names: The list of the columns whose distinct values should be
                returned.
            table_name: The name of the table from which the column should be loaded.
                Defaults to None.

        Returns:
            The distinct values of the requested column as a mapping from col name to
            a series of distinct values.

        Raises:
            ValueError: If the table name provided does not exist.
            ValueError: If the data is multi-table but no table name provided.
        """
        if table_name is None and self.multi_table:
            raise ValueError("No table name provided for multi-table datasource.")
        elif table_name and self.multi_table:
            self._validate_table_name(table_name)
            df = cast(pd.DataFrame, self.get_data(table_name=table_name))
        else:
            # if single-table datasource
            df = cast(pd.DataFrame, self.get_data(table_name=table_name))
        return {col: df[col].unique() for col in col_names}

    def get_column(
        self, col_name: str, table_name: Optional[str] = None, **kwargs: Any
    ) -> pd.Series:
        """Loads and returns single column from Excel dataset.

        Args:
            col_name: The name of the column which should be loaded.
            table_name: The name of the table from which the column should be loaded.
                Defaults to None.

        Returns:
            The column request as a series.

        Raises:
            ValueError: If the table name provided does not exist.
            ValueError: If the data is multi-table but no table name provided.
        """
        if table_name is None and self.multi_table:
            raise ValueError("No table name provided for multi-table datasource.")
        elif table_name and self.multi_table:
            self._validate_table_name(table_name)
            df = cast(pd.DataFrame, self.get_data(table_name=table_name))
        else:
            # if single-table datasource
            df = cast(pd.DataFrame, self.get_data(table_name=table_name))
        return df[col_name]

    def get_dtypes(self, table_name: Optional[str] = None, **kwargs: Any) -> _Dtypes:
        """Loads and returns the columns and column types of the Excel dataset.

        Returns:
            A mapping from column names to column types.

        Raises:
            ValueError: If the table name provided does not exist.
            ExcelSourceError: If no table name provided.
        """
        if table_name:
            df = cast(pd.DataFrame, self.get_data(table_name=table_name))
            dtypes = self._get_data_dtypes(df)
            # We use this function to generate the schema,
            # no point to keep the data loaded in memory,
            # so delete the df.
            del df
        elif not self.multi_table:
            df = cast(pd.DataFrame, self.get_data())
            dtypes = self._get_data_dtypes(df)
            # We use this function to generate the schema,
            # no point to keep the data loaded in memory,
            # so delete the df.
            del df
        else:
            raise ExcelSourceError(
                "No table_name provided for multi_table ExcelSource. Please provide"
                "a table name to retrieve the columns and column types for this "
                "Excel dataset."
            )
        return dtypes

    def __len__(self) -> int:
        if self._data_is_loaded:
            return len(self.data)
        elif not self.multi_table:
            data = self.get_data()
            assert data is not None  # nosec[assert_used]
            self.data = data
            return len(self.data)

        raise ValueError("Can't ascertain length of multi-table Excel dataset.")

    @property
    def multi_table(self) -> bool:
        """Attribute to specify whether the datasource is multi table."""
        if self.sheet_name is not None:
            return not isinstance(self.sheet_name, str) and len(self.sheet_name) > 1

        return len(self.table_names) > 1



"""Handling incoming tasks for a Pod."""
from __future__ import annotations

from typing import Iterable, Mapping, Optional, Union

from grpc import RpcError

from bitfount.federated.exceptions import PodConnectFailedError
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.transport.base_transport import Handler, _BaseMailbox
from bitfount.federated.transport.message_service import (
    _BitfountMessageType,
    _MessageService,
)

DEFAULT_KEY_EXCHANGE_TIMEOUT: int = 2 * 60

logger = _get_federated_logger(__name__)


class _PodMailbox(_BaseMailbox):
    """Handling of incoming training tasks from modellers."""

    def __init__(
        self,
        pod_name: str,
        mailbox_id: str,
        message_service: _MessageService,
        handlers: Optional[
            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
        ] = None,
    ):
        """Creates a new pod mailbox.

        Note that the preferred way to get a new PodMailbox is by calling
        PodMailbox.connect_pod() which will instantiate the correct PodMailbox
        for you.

        Args:
            pod_name: The name of the pod.
            mailbox_id: The mailbox ID for this pod on the message service.
            message_service: The message service to use.
            handlers: Optional. A set of handlers to initialise with.
        """
        super().__init__(
            mailbox_id=mailbox_id, message_service=message_service, handlers=handlers
        )
        self.pod_name = pod_name
        self.pod_identifier = f"{self.message_service.username}/{self.pod_name}"

    @classmethod
    async def connect_pod(
        cls, pod_name: str, message_service: _MessageService
    ) -> _PodMailbox:
        """Registers with message service and created mailbox.

        Returns:
            The created mailbox.

        Raises:
            RuntimeError: If a GRPC error occurs or the message service is
                          unable to connect the pod.
        """
        logger.info(f"Connecting to messaging service as: {pod_name}")
        try:
            mailbox_id = await message_service.connect_pod(pod_name)
        except (RpcError, PodConnectFailedError) as err:
            logger.critical(
                "Error occurred when trying to call PodConnect on messaging service"
            )
            raise PodConnectFailedError(
                f"Failed to connect to messaging service "
                f"as pod: {pod_name}. Error: {err}"
            )

        pod_mailbox = cls(
            pod_name=pod_name, mailbox_id=mailbox_id, message_service=message_service
        )

        return pod_mailbox



"""Monitor module and per-task context handling."""
from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
import logging
import threading
from typing import TYPE_CHECKING, Dict, Final, Generator, Optional, Union

from requests import RequestException

from bitfount.federated.monitoring.exceptions import (
    ExistingMonitorModuleError,
    NoMonitorModuleError,
)
from bitfount.federated.monitoring.types import (
    AdditionalMonitorMessageTypes,
    MonitorRecordPrivacy,
    ProgressCounterDict,
)
from bitfount.federated.transport.message_service import _BitfountMessageType
from bitfount.types import _JSONDict

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub
    from bitfount.hub.types import _MonitorPostJSON

_logger = logging.getLogger(__name__)


@dataclass
class _MonitorPostContent:
    """Python representation of a monitor POST body.

    Before sending the request use the `json()` method to convert it to the required
    format for sending.
    """

    task_id: str
    sender_id: str
    timestamp: datetime
    privacy: MonitorRecordPrivacy
    type: Union[_BitfountMessageType, AdditionalMonitorMessageTypes]

    # Not required fields
    recipient_id: Optional[str] = None
    message: Optional[str] = None
    metadata: Optional[_JSONDict] = None
    progress: Optional[Dict[str, ProgressCounterDict]] = None
    resource_usage: Optional[Dict[str, float]] = None

    def json(self) -> _MonitorPostJSON:
        """Converts this object into the equivalent JSON representation."""
        # Extract type as a string dependent on form
        type_str: str
        if isinstance(self.type, _BitfountMessageType):
            type_str = self.type.name
        else:  # AdditionalMonitorMessageTypes
            type_str = self.type.value

        json_dict: _MonitorPostJSON = {
            "senderId": self.sender_id,
            "timestamp": self.timestamp.isoformat(),
            "privacy": self.privacy.value,
            "type": type_str,
            "taskId": self.task_id,
        }

        # Not required fields
        if self.recipient_id:
            json_dict["recipientId"] = self.recipient_id
        if self.message:
            json_dict["message"] = self.message
        if self.metadata:
            json_dict["metadata"] = self.metadata
        if self.progress:
            json_dict["progress"] = self.progress
        if self.resource_usage:
            json_dict["resourceUsage"] = self.resource_usage

        return json_dict


class _MonitorModule:
    """Class for interacting with monitoring service endpoints."""

    def __init__(
        self,
        hub: BitfountHub,
        task_id: str,
        sender_id: str,
        recipient_id: Optional[str] = None,
    ):
        """Create a new monitoring service interface.

        Args:
            hub: The hub to send monitoring details to.
            task_id: The ID of the task that monitor updates are associated with.
            sender_id: The mailbox ID of this end of the task (i.e. worker mailbox
                ID for the worker, modeller mailbox ID for the modeller).
            recipient_id: Optional. The mailbox ID for the other end of the task
                if appropriate.
        """
        self._hub = hub
        self._task_id = task_id
        self._sender_id = sender_id
        self._recipient_id = recipient_id

    @property
    def task_id(self) -> str:
        """Task ID for this monitor module."""
        return self._task_id

    def _send_to_monitor_service(self, update: _MonitorPostContent) -> None:
        """Send updates to the monitoring service.

        Logs out if an error is encountered.
        """
        try:
            self._hub.send_monitor_update(update.json())
        except RequestException as exc:
            _logger.warning(
                f"Unable to send monitoring update of type {update.type.value};"
                f" error was: {str(exc)}"
            )

    def send_to_monitor_service(
        self,
        event_type: Union[_BitfountMessageType, AdditionalMonitorMessageTypes],
        privacy: MonitorRecordPrivacy,
        message: Optional[str] = None,
        metadata: Optional[_JSONDict] = None,
        progress: Optional[Dict[str, ProgressCounterDict]] = None,
        resource_usage: Optional[Dict[str, float]] = None,
    ) -> None:
        """Send updates to the monitoring service.

        Logs out if an error is encountered.
        """
        # Set timestamp to current UTC
        timestamp = datetime.now(timezone.utc)

        self._send_to_monitor_service(
            _MonitorPostContent(
                task_id=self._task_id,
                sender_id=self._sender_id,
                timestamp=timestamp,
                privacy=privacy,
                type=event_type,
                recipient_id=self._recipient_id,
                message=message,
                metadata=metadata,
                progress=progress,
                resource_usage=resource_usage,
            )
        )


# ###################################### #
# Global monitor module maintenance code #
# ###################################### #
# Global monitor module instance, to be set when task ID is known,
# removed at end of task
_MONITOR_MODULE: Optional[_MonitorModule] = None
# Lock to ensure atomic creation/access to monitoring module
# We use a reentrant lock here because we're happy for a single thread to acquire
# the lock multiple times (as these methods are all sync they can't interfere with
# themselves/each other).
# NOTE: If we make these async at any point then this lock will need to be revisited
#       as we will lose the guarantee that a single thread is safe from trying to
#       acquire it from multiple locations "simultaneously".
_MONITOR_MODULE_LOCK: Final[threading.RLock] = threading.RLock()


def _set_task_monitor(
    hub: BitfountHub, task_id: str, sender_id: str, recipient_id: Optional[str] = None
) -> None:
    """Sets the global monitor module instance.

    Raises:
        ExistingMonitorModuleError: If a monitor module is already set.
    """
    global _MONITOR_MODULE
    with _MONITOR_MODULE_LOCK:
        if _MONITOR_MODULE is not None:
            raise ExistingMonitorModuleError(
                f"A monitor module already exists,"
                f" associated with {_MONITOR_MODULE.task_id}"
            )
        else:
            _MONITOR_MODULE = _MonitorModule(hub, task_id, sender_id, recipient_id)


def _unset_task_monitor() -> None:
    """Removes the global monitor module instance."""
    global _MONITOR_MODULE
    with _MONITOR_MODULE_LOCK:
        _MONITOR_MODULE = None


def _get_task_monitor() -> _MonitorModule:
    """Returns the current monitor module instance.

    Raises:
        NoMonitorModuleError: If no monitor module has been created yet.
    """
    global _MONITOR_MODULE
    with _MONITOR_MODULE_LOCK:
        if _MONITOR_MODULE is None:
            raise NoMonitorModuleError(
                "No monitor module available,"
                " please create one using task_monitor_context()."
            )
        else:
            return _MONITOR_MODULE


@contextmanager
def task_monitor_context(
    hub: BitfountHub, task_id: str, sender_id: str, recipient_id: Optional[str] = None
) -> Generator[None, None, None]:
    """Context manager that handles monitor module instance creation and removal.

    Used to provide "task context" in which updates can be sent to the monitor
    service which are related to that task.
    """
    _set_task_monitor(hub, task_id, sender_id, recipient_id)
    try:
        yield
    finally:
        _unset_task_monitor()


def task_status_update(
    message: str,
    privacy: MonitorRecordPrivacy = MonitorRecordPrivacy.ALL_PARTICIPANTS,
    metadata: Optional[_JSONDict] = None,
    progress: Optional[Dict[str, ProgressCounterDict]] = None,
    resource_usage: Optional[Dict[str, float]] = None,
) -> None:
    """Send a task status update to the monitor service.

    Args:
        message: The update message to send.
        privacy: The privacy level of the status update. Defaults to all participants.
        metadata: Optional. Any metadata about the status update.
        progress: Optional. Any progress counter information about the status update.
        resource_usage: Optional. Any resource usage information about the status
            update.
    """
    try:
        with _MONITOR_MODULE_LOCK:
            monitor = _get_task_monitor()
            monitor.send_to_monitor_service(
                event_type=AdditionalMonitorMessageTypes.TASK_STATUS_UPDATE,
                privacy=privacy,
                message=message,
                metadata=metadata,
                progress=progress,
                resource_usage=resource_usage,
            )
        _logger.debug("Sent task status update")
    except NoMonitorModuleError as exc:
        _logger.warning(f"Unable to send task status update: {str(exc)}")


def task_config_update(
    task_config: _JSONDict,
    privacy: MonitorRecordPrivacy = MonitorRecordPrivacy.OWNER_MODELLER,
) -> None:
    """Send the final task config to the monitor service.

    Args:
        task_config: The task configuration as a JSON-compatible dictionary.
        privacy: The privacy level of the update. Defaults to all participants.
    """
    try:
        with _MONITOR_MODULE_LOCK:
            monitor = _get_task_monitor()
            monitor.send_to_monitor_service(
                event_type=AdditionalMonitorMessageTypes.TASK_CONFIG,
                privacy=privacy,
                metadata=task_config,
            )
        _logger.debug("Sent task config update")
    except NoMonitorModuleError as exc:
        _logger.warning(f"Unable to send task config update: {str(exc)}")



"""Binary (two reference arguments) transformations.

This module contains the base class and concrete classes for binary transformations,
those that can take two reference arguments (i.e. column or transformation names).
"""
from __future__ import annotations

from typing import Any, Union

import attr

from bitfount.transformations.base_transformation import Transformation
from bitfount.utils import delegates


@delegates()
@attr.dataclass(kw_only=True)
class BinaryOperation(Transformation):
    """Base two-arg transformation.

    The base abstract class for all Binary Operation Transformations.

    Args:
        arg1: The first argument
        arg2: The second argument
    """

    arg1: Any
    arg2: Any


@delegates()
@attr.dataclass(kw_only=True)
class NumericBinaryOperation(BinaryOperation):
    """Base two-arg operation involving numbers.

    This class represents any BinaryOperation where arg2 can be numeric or a
    second column name such as addition, multiplication, etc.

    Args:
        arg1: The first argument (column name).
        arg2: The second argument (column name or numeric value).
    """

    arg1: str
    arg2: Union[float, str]


@delegates()
@attr.dataclass(kw_only=True)
class AdditionTransformation(NumericBinaryOperation):
    """Represents the addition of two columns or of a constant to a column."""

    _registry_name = "add"


@delegates()
@attr.dataclass(kw_only=True)
class SubtractionTransformation(NumericBinaryOperation):
    """Column subtracting transformation.

    Represents the subtraction of one column from another or of a constant from
    a column.
    """

    _registry_name = "subtract"


@delegates()
@attr.dataclass(kw_only=True)
class MultiplicationTransformation(NumericBinaryOperation):
    """Represents the multiplication of two columns or of a column and a constant."""

    _registry_name = "multiply"


@delegates()
@attr.dataclass(kw_only=True)
class DivisionTransformation(NumericBinaryOperation):
    """Column division transformation.

    Represents the division of one column by another or of one column by a constant.
    """

    _registry_name = "divide"


@delegates()
@attr.dataclass(kw_only=True)
class ComparisonTransformation(NumericBinaryOperation):
    """Represents the comparison between two columns or of one column and a constant.

    The resulting output should be:
        - -1 if arg1 < arg2
        - 0 if arg1 == arg2
        - +1 if arg1 > arg2.
    """

    _registry_name = "compare"



"""File containing mock objects for the purposes of testing."""
from __future__ import annotations

import asyncio
from collections.abc import Awaitable
import dataclasses
import inspect
from inspect import Parameter
import logging
import multiprocessing
from multiprocessing.managers import DictProxy, ListProxy, SyncManager
import queue
import threading
import time
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Collection,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from unittest.mock import AsyncMockMixin, NonCallableMock, create_autospec
import uuid

from grpc import RpcError, StatusCode
from typing_extensions import TypeAlias

from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
    _MessageService,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    BitfountMessage as GrpcBitfountMessage,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    BitfountTasks,
    BlobStorageData,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    CommunicationDetails as GrpcCommunicationDetails,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    LargeStorageRequest,
    PodData,
    SuccessResponse,
)
from bitfount.federated.transport.protos.messages_pb2 import Acknowledgement
from bitfount.federated.transport.protos.messages_pb2_grpc import MessageServiceStub
from bitfount.federated.transport.types import CommunicationDetails

logger = logging.getLogger(__name__)


class LocalMessageServiceSharedQueues:
    """Class that simulates message mailboxes with asyncio.Queues."""

    def __init__(self) -> None:
        """Create new LocalMessageServiceSharedQueues instance."""
        self._queues: Dict[str, queue.SimpleQueue[_BitfountMessage]] = {}

    def send_message(self, mailbox_id: str, message: _BitfountMessage) -> None:
        """Sends a message to the queue for mailbox_id."""
        self._queues[mailbox_id].put_nowait(message)

    async def get_message(self, mailbox_id: str) -> _BitfountMessage:
        """Awaits on a message to pull from queue for mailbox_id."""
        while True:
            try:
                message = self._queues[mailbox_id].get_nowait()
                logger.debug(
                    f"Retrieved message ({message.message_type}) from {mailbox_id}"
                )
                return message
            except queue.Empty:
                await asyncio.sleep(0)

    def add_mailbox(self, mailbox_id: str) -> None:
        """Creates a new queue for mailbox_id."""
        self._queues[mailbox_id] = queue.SimpleQueue()


class LocalMessageService(_MessageService):
    """A class that implements the MessageService using asyncio.Queues.

    Has the same interface as MessageService but is designed for local testing.
    """

    # noinspection PyMissingConstructor
    def __init__(
        self,
        username: str,
        shared_queues: LocalMessageServiceSharedQueues,
        modeller_mailbox_id: Optional[str] = None,
        pod_ids: Optional[Collection[str]] = None,
        task_id: Optional[str] = None,
    ):
        """Create new LocalMessageService.

        Args:
            username: The username of the account "running" this message service.
            shared_queues: The shared queues to act as the mailboxes.
            modeller_mailbox_id: Optional. The modeller's mailbox ID to simulate with.
            pod_ids: Optional. The group of pod_ids to simulate with.
            task_id: Optional. The task ID to simulate with. If not provided, one
                will be generated.
        """
        # We don't have a BitfountSession to keep track of username with, so we
        # need it manually.
        self._username = username
        self._shared_queues = shared_queues
        self._modeller_mailbox_id = modeller_mailbox_id
        self._pod_ids = pod_ids

        self._task_id: str
        if task_id is None:
            self._task_id = uuid.uuid4().hex
        else:
            self._task_id = task_id

    @property
    def username(self) -> str:
        """Username running this message service."""
        return self._username

    @property
    def modeller_mailbox_id(self) -> str:
        """The modeller_mailbox_id to use for this message service."""
        if self._modeller_mailbox_id:
            return self._modeller_mailbox_id
        else:
            raise ValueError("No modeller_mailbox_id set.")

    @property
    def worker_mailbox_ids(self) -> Dict[str, str]:
        """A mapping of pod identifiers to worker_mailbox_ids.

        Deterministically generated so that they can be accessed consistently
        across all LocalMessageService instances using the same queue backend.
        """
        if not self._pod_ids:
            raise ValueError("No worker_mailbox_ids set.")
        return {
            pod_identifier: self._get_worker_mailbox_id(pod_identifier)
            for pod_identifier in self._pod_ids
        }

    async def connect_pod(self, pod_name: str) -> str:
        """See parent for more information."""
        # TODO: [BIT-960] Currently this is just hardcoded to return the pod_name
        #       (which is what the mailbox ID will actually be). [BIT-960] will have
        #       the PodConnect method actually return the generated mailbox ID so
        #       that if the approach changes in future it only needs to change on
        #       the message service side. At that point this should return the
        #       generated mailbox ID instead.
        pod_mailbox_id = pod_name
        # Create queue for pod
        self._shared_queues.add_mailbox(pod_mailbox_id)
        return pod_mailbox_id

    async def setup_communication_with_pods(
        self, tasks_per_pod: Mapping[str, bytes]
    ) -> CommunicationDetails:
        """See parent for more information."""
        # Create queues for modeller and for workers
        self._shared_queues.add_mailbox(self.modeller_mailbox_id)

        for worker_mailbox_id in self.worker_mailbox_ids.values():
            self._shared_queues.add_mailbox(worker_mailbox_id)

        return CommunicationDetails(
            self.modeller_mailbox_id, self.worker_mailbox_ids, self._task_id
        )

    async def poll_for_messages(
        self,
        mailbox_id: str,
        stop_event: threading.Event,
    ) -> AsyncGenerator[_BitfountMessage, None]:
        """See parent for more information."""
        # Check stop condition at start of loop
        while not stop_event.is_set():
            # Check stop condition whilst waiting for message
            stop_fut = asyncio.get_running_loop().run_in_executor(None, stop_event.wait)
            get_message_task = asyncio.create_task(
                self._shared_queues.get_message(mailbox_id)
            )

            awaitables: List[Awaitable] = [get_message_task, stop_fut]
            done, pending = await asyncio.wait(
                awaitables, return_when=asyncio.FIRST_COMPLETED
            )

            if get_message_task in done:
                yield get_message_task.result()
            else:  # stop_event is set
                get_message_task.cancel()
                return

    async def send_message(
        self, message: _BitfountMessage, already_packed: bool = False
    ) -> SuccessResponse:
        """See parent for more information."""
        logger.debug(
            f"Sending message {message.message_type} to {message.recipient_mailbox_id}"
        )
        self._shared_queues.send_message(message.recipient_mailbox_id, message)
        return cast(SuccessResponse, create_autospec(SuccessResponse, instance=True))

    @staticmethod
    def _get_worker_mailbox_id(pod_identifier: str) -> str:
        """Deterministically generates a worker mailbox ID from a pod identifier."""
        return f"{pod_identifier}_worker"


class GRPCStubMock(MessageServiceStub):
    """Multiprocess friendly GRPC Stub Fake.

    This mocks the behaviour of the message service.
    It can handle a single training job.
    It assumes the happy path is followed, the only 'error' it handles is
    when no messages are available.
    """

    # noinspection PyMissingConstructor
    def __init__(
        self,
        tokens_to_usernames: Mapping[str, str],
        manager: SyncManager,
        get_message_timeout: float = 0.1,
    ):
        self.tokens_to_usernames = tokens_to_usernames
        self.manager = manager
        self.get_message_timeout = get_message_timeout
        self.queue_id_counter = manager.Value("i", 1)
        self.large_object_id_counter = manager.Value("i", 1)

        self.queues: DictProxy[str, ListProxy[GrpcBitfountMessage]] = manager.dict()

        self.lock = multiprocessing.Lock()

    @staticmethod
    def _get_token(metadata: Sequence[Tuple[str, str]]) -> str:
        """Retrieve token from metadata."""
        return metadata[0][1]

    async def PodConnect(
        self,
        data: PodData,
        /,
        *,
        metadata: Sequence[Tuple[str, Any]],
        timeout: Optional[float] = None,
    ) -> SuccessResponse:
        """Fakes PodConnect behaviour."""
        pod_id = self.tokens_to_usernames[self._get_token(metadata)]
        pod_queue_name = pod_id.replace("/", "-")
        with self.lock:
            self.queues[pod_queue_name] = self.manager.list()

        return SuccessResponse()

    async def SetupTaskMailboxes(
        self,
        data: BitfountTasks,
        /,
        *,
        metadata: Sequence[Tuple[str, Any]],
        timeout: Optional[float] = None,
    ) -> GrpcCommunicationDetails:
        """Fakes SetupTaskMailboxes behaviour."""
        modeller_name = self.tokens_to_usernames[self._get_token(metadata)]

        task_id: str = uuid.uuid4().hex

        with self.lock:
            modeller_queue_id = self.queue_id_counter.value
            self.queues[
                f"{modeller_name}-{self.queue_id_counter.value}"
            ] = self.manager.list()
            self.queue_id_counter.value += 1

            pod_mailbox_ids = {}
            for task in data.tasks:
                print(f"Queues are: {self.queues}")
                namespace, name = task.podIdentifier.split("/")
                # Pod queue creation
                print(f"making queue: '{namespace}-{self.queue_id_counter.value}'")
                self.queues[
                    f"{namespace}-{self.queue_id_counter.value}"
                ] = self.manager.list()
                pod_mailbox_ids[task.podIdentifier] = str(self.queue_id_counter.value)
                self.queue_id_counter.value += 1

            for task in data.tasks:
                namespace, name = task.podIdentifier.split("/")
                general_mailbox_name = f"{namespace}-{name}"
                print(f"MOCK MAILBOXES: {pod_mailbox_ids}")
                self.queues[general_mailbox_name].append(
                    GrpcBitfountMessage(
                        messageType=_BitfountMessageType.JOB_REQUEST.value,
                        body=task.encryptedTask,
                        sender=modeller_name,
                        senderMailboxId=str(modeller_queue_id),
                        recipient=task.podIdentifier,
                        recipientMailboxId=general_mailbox_name,
                        podMailboxIds=pod_mailbox_ids,
                        taskId=task_id,
                    )
                )

        print(f"POD MAILBOXES: {pod_mailbox_ids}")
        return GrpcCommunicationDetails(
            mailboxId=str(modeller_queue_id),
            podMailboxIds=pod_mailbox_ids,
            taskId=task_id,
        )

    async def SendBitfountMessage(
        self,
        data: GrpcBitfountMessage,
        /,
        *,
        metadata: Sequence[Tuple[str, Any]],
        timeout: Optional[float] = None,
    ) -> SuccessResponse:
        """Fakes SendBitfountMessage behaviour."""
        sender_name = self.tokens_to_usernames[self._get_token(metadata)]
        recipient_username = data.recipient

        if "/" in data.recipient:
            # It's for a pod
            recipient_username, pod_name = data.recipient.split("/")

        with self.lock:
            self.queues[f"{recipient_username}-{data.recipientMailboxId}"].append(
                GrpcBitfountMessage(
                    messageType=data.messageType,
                    body=data.body,
                    recipient=recipient_username,
                    recipientMailboxId=data.recipientMailboxId,
                    sender=sender_name,
                    senderMailboxId=data.senderMailboxId,
                    taskId=data.taskId,
                )
            )
        return SuccessResponse()

    async def GetBitfountMessage(
        self,
        data: GrpcCommunicationDetails,
        /,
        *,
        metadata: Sequence[Tuple[str, Any]],
        timeout: Optional[float] = None,
    ) -> GrpcBitfountMessage:
        """Fakes GetBitfountMessage behaviour."""
        user_name = self.tokens_to_usernames[self._get_token(metadata)]

        if "/" in user_name:
            # It's a pod, we need to split out the pod name from the username
            user_name = user_name.split("/")[0]

        self.lock.acquire()
        mailbox = self.queues[f"{user_name}-{data.mailboxId}"]

        if len(mailbox) == 0:
            self.lock.release()
            time.sleep(self.get_message_timeout)
            self.lock.acquire()

            if len(mailbox) == 0:
                self.lock.release()

                # Create an RpcError, manually setting the code property as this
                # class is normally constructed in C code.
                error = RpcError()
                error.code = lambda: StatusCode.NOT_FOUND  # type: ignore[assignment] # Reason: see comment # noqa: B950

                raise error

        message = mailbox[0]
        self.lock.release()

        return message

    async def AcknowledgeMessage(
        self,
        data: Acknowledgement,
        /,
        *,
        metadata: Sequence[Tuple[str, Any]],
        timeout: Optional[float] = None,
    ) -> SuccessResponse:
        """Fakes AcknowledgeMessage behaviour."""
        user_name = self.tokens_to_usernames[self._get_token(metadata)]
        if "/" in user_name:
            # It's a pod, we need to split out the pod name from the username
            user_name = user_name.split("/")[0]

        with self.lock:
            self.queues[f"{user_name}-{data.mailboxId}"].pop(0)

        return SuccessResponse()

    async def GetLargeObjectStorage(
        self,
        data: LargeStorageRequest,
        /,
        *,
        metadata: Sequence[Tuple[str, Any]],
        timeout: Optional[float] = None,
    ) -> BlobStorageData:
        """Get URLs for upload/download of a large message."""
        # Locking just to ensure that we have a unique counter value
        with self.lock:
            print(
                f"Large storage {self.large_object_id_counter.value} "
                f"taken by: {self.tokens_to_usernames[self._get_token(metadata)]} "
                f"(pod: {data.podName})"
            )
            result = BlobStorageData(
                uploadUrl=f"https://test-message-service-"
                f"external.s3.eu-west-2.amazonaws.com/"
                f"upload?large-object-id="
                f"{self.large_object_id_counter.value}",
                downloadUrl=f"https://test-message-service-"
                f"external.s3.eu-west-2.amazonaws.com/"
                f"download?large-object-id="
                f"{self.large_object_id_counter.value}",
                uploadFields={"key": "some/key", "bucket": "some-bucket"},
            )
            self.large_object_id_counter.value += 1
            return result


# Type var for AsyncIteratorMock
T = TypeVar("T")


class AsyncIteratorMock(AsyncIterator, Generic[T]):
    """A mocked async iterator that will iterate a sync iterable."""

    def __init__(self, iterable: Iterable[T]) -> None:
        self.iterator: Iterator[T] = iter(iterable)

    def __aiter__(self) -> AsyncIteratorMock[T]:
        return self

    async def __anext__(self) -> T:
        # Iterate through the iterable, raising exceptions if they are present
        try:
            result = next(self.iterator)
            if self._is_exception(result):
                raise cast(BaseException, result)
            return result
        except StopIteration:
            pass
        raise StopAsyncIteration

    @staticmethod
    def _is_exception(obj: Any) -> bool:
        return (
            isinstance(obj, BaseException)
            or isinstance(obj, type)
            and issubclass(obj, BaseException)
        )


DataclassMock: TypeAlias = NonCallableMock


def create_dataclass_mock(
    dataclass_cls_or_instance: Union[Any, Type[Any]]
) -> DataclassMock:
    """Creates a mock whose spec matches the attributes of target dataclass.

    The actual mock is a NonCallableMock instance as it makes no sense to call
    a dataclass instance.
    """
    if not dataclasses.is_dataclass(dataclass_cls_or_instance):
        raise TypeError("create_dataclass_mock should only be used on dataclasses.")

    # To force create_autospec to find all the attributes (including those that
    # don't have default values) we need to instantiate a dataclass instance to
    # get everything set. If we already have an instance this is fine, otherwise
    # we need to work out (and pass) the number of expected args.
    dataclass_instance: Any
    if inspect.isclass(dataclass_cls_or_instance):
        # Work out and create instance from class
        dataclass_cls: Type[Any] = dataclass_cls_or_instance
        sig = inspect.signature(dataclass_cls)
        params = sig.parameters.values()
        # Find params that _can_ be supplied positionally
        args = [
            p
            for p in params
            if p.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
        ]
        # Find params that don't have default values (Parameter.empty denotes no
        # default value)
        needed_args = [p for p in args if p.default == Parameter.empty]
        false_args = [None] * len(needed_args)
        dataclass_instance = dataclass_cls(*false_args)
    else:
        # Use instance directly
        dataclass_instance = cast(Any, dataclass_cls_or_instance)

    dataclass_mock: DataclassMock = create_autospec(
        spec=dataclass_instance, instance=True
    )

    return dataclass_mock


class AwaitableMock(AsyncMockMixin, NonCallableMock, Awaitable):
    """A non-callable async mock which is also directly awaitable.

    Useful for mocking out things like Task or Future instances.

    It implements __await__ to yield and return `None` (which gets around
    compatibility issues with asyncio not being able to await a Mock subclass
    directly) and to record the call to __await__.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    def __await__(self) -> Generator[None, None, None]:
        self.await_count += 1
        yield None



"""Tests for backends/pytorch/types.py."""
import torch

from bitfount.backends.pytorch.types import _AdaptorForPyTorchTensor
from tests.utils.helper import backend_test, unit_test


@backend_test
@unit_test
class TestAdaptorForPyTorchTensor:
    """Tests for _AdaptorForPyTorchTensor."""

    def test_squeeze(self) -> None:
        """Tests squeeze method."""
        adaptor = _AdaptorForPyTorchTensor(torch.tensor([[1, 2, 3]]))
        new_adaptor = adaptor.squeeze()
        assert isinstance(new_adaptor, _AdaptorForPyTorchTensor)
        # Testing that shape has been reduced from 2 dimensions to 1
        assert len(adaptor.torchtensor.shape) == 2
        assert len(new_adaptor.torchtensor.shape) == 1

    def test_squeeze_axis(self) -> None:
        """Tests squeeze method with axis."""
        adaptor = _AdaptorForPyTorchTensor(torch.zeros(2, 1, 2, 1, 2))
        assert list(adaptor.torchtensor.shape) == [2, 1, 2, 1, 2]

        # Shape does not change because the tensor has not been squeezed
        # in the right dimension
        new_adaptor = adaptor.squeeze(axis=0)
        assert list(new_adaptor.torchtensor.shape) == [2, 1, 2, 1, 2]

        # Shape only changes in dimension 1
        new_adaptor = adaptor.squeeze(axis=1)
        assert list(new_adaptor.torchtensor.shape) == [2, 2, 1, 2]



"""Contains classes for marking differential privacy on models."""
from __future__ import annotations

from dataclasses import dataclass, field
import inspect
from types import MappingProxyType
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

import desert
import marshmallow
from marshmallow import Schema as MarshmallowSchema
from marshmallow import fields
from marshmallow.validate import OneOf
from marshmallow_union import Union as M_Union

from bitfount.federated.exceptions import DPParameterError
from bitfount.federated.logging import _get_federated_logger
from bitfount.types import (
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    _BaseSerializableObjectMixIn,
    _StrAnyDict,
)

logger = _get_federated_logger(__name__)

_DEFAULT_ALPHAS: List[float] = [1 + x / 10.0 for x in range(1, 100)] + list(
    range(12, 64)
)
_DEFAULT_DELTA: float = 1e-6
_ALLOWED_LOSS_REDUCTIONS: Tuple[str, str] = ("mean", "sum")

# The mutable underlying dict that holds the registry information
_registry: Dict[str, Type[_BaseDPConfig]] = {}
# The read-only version of the registry that is allowed to be imported
registry: Mapping[str, Type[_BaseDPConfig]] = MappingProxyType(_registry)


@dataclass
class _BaseDPConfig:
    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
            _registry[cls.__name__] = cls


@dataclass
class DPModellerConfig:
    """Modeller configuration options for Differential Privacy.

    :::info

    `epsilon` and `delta` are also set by the Pods involved in the task and
    take precedence over the values supplied here.

    :::

    Args:
        epsilon: The maximum epsilon value to use.
        max_grad_norm: The maximum gradient norm to use. Defaults to 1.0.
        alphas: The alphas to use. Defaults to floats from 1.1 to 63.0 (inclusive) with
            increments of 0.1 up to 11.0 followed by increments of 1.0 up to 63.0.
            Note that none of the alphas should be equal to 1.
        delta: The target delta to use. Defaults to 1e-6.
        loss_reduction: The loss reduction to use. Available options are "mean" and
            "sum". Defaults to "mean".
        auto_fix: Whether to automatically fix the model if it is not DP-compliant.
            Currently, this just converts all `BatchNorm` layers to `GroupNorm`.
            Defaults to True.

    Raises:
        ValueError: If loss_reduction is not one of "mean" or "sum".
    """

    # DP directly related options
    epsilon: float
    max_grad_norm: Union[float, List[float]] = 1.0
    # TODO: [BIT-1614] Re-enable user-provided noise_multiplier after the post-update
    #       privacy violation issue is resolved.
    # noise_multiplier: float = 0.4
    alphas: List[float] = field(default_factory=lambda: _DEFAULT_ALPHAS)
    delta: float = _DEFAULT_DELTA
    loss_reduction: Literal["mean", "sum"] = desert.field(  # type: ignore[assignment] # Reason: desert.field is intentional # noqa: B950
        marshmallow.fields.String(validate=OneOf(_ALLOWED_LOSS_REDUCTIONS)),
        default="mean",
    )

    # Other options
    auto_fix: bool = True
    fields_dict = {
        "epsilon": fields.Float(),
        "max_grad_norm": M_Union([fields.Float(), fields.List(fields.Float())]),
        "alphas": fields.List(fields.Float()),
        "delta": fields.Float(),
        "loss_reduction": fields.String(),
        "auto_fix": fields.Bool(),
    }  # type:  _StrAnyDict
    nested_fields = {}  # type:Dict[str, _StrAnyDict]

    def __post_init__(self) -> None:
        # Validate loss_reduction
        if self.loss_reduction not in _ALLOWED_LOSS_REDUCTIONS:
            raise ValueError(
                f"loss_reduction must be one of {_ALLOWED_LOSS_REDUCTIONS}, "
                f'not "{self.loss_reduction}".'
            )
        if any(alpha == 1 for alpha in self.alphas):
            raise DPParameterError(
                "You are trying to run a DP-task with an alpha value of 1, "
                "which is not allowed. Please update your alpha values and "
                "try again."
            )


@dataclass
class DPPodConfig(_BaseDPConfig):
    """Pod configuration options for Differential Privacy.

    Primarily used as caps and bounds for what options may be set by the modeller.

    Args:
        epsilon: The maximum epsilon value to use.
        delta: The maximum target delta to use. Defaults to 1e-6.
    """

    epsilon: float
    delta: float = _DEFAULT_DELTA
    fields_dict: _StrAnyDict = field(
        default_factory=lambda: {
            "epsilon": fields.Float(),
            "delta": fields.Float(),
        }
    )


class _DifferentiallyPrivate(_BaseSerializableObjectMixIn):
    """Marks that the model supports differential privacy.

    This class itself does not handle the implementation details of DP (as that
    will differ on a per-model/per-library basis) but captures the configuration
    details for DP to enable implementing models to make use of it.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "_dp_config": fields.Nested(
            desert.schema(DPModellerConfig),
            allow_none=True,
            data_key="dp_config",
            dump_only=True,
        ),
        "dp_config": fields.Nested(
            desert.schema(DPModellerConfig),
            allow_none=True,
            data_key="dp_config",
            load_only=True,
        ),
    }
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}

    def __init__(
        self,
        dp_config: Optional[Union[DPModellerConfig, Mapping[str, Any]]] = None,
        **kwargs: Any,
    ):
        """Stores the Differential Privacy configuration for this model.

        Args:
            dp_config: The Differential Privacy configuration to use. Can be either
                a `DPModellerConfig` instance or a string-keyed mapping.
            **kwargs: Other keyword arguments to be passed up the inheritance
                hierarchy.
        """
        self.class_name = type(self).__name__
        # Capture DP config options, converting if needed
        if not isinstance(dp_config, DPModellerConfig):
            dp_config = self._convert_to_dpconfig(dp_config)
        self._dp_config: Optional[DPModellerConfig] = dp_config

        if self._dp_config:
            logger.info(
                f"Model is differentially private with settings: {self._dp_config}"
            )
        else:
            logger.info("No differential privacy settings provided.")

        super().__init__(**kwargs)

    @staticmethod
    def _convert_to_dpconfig(
        dict_config: Optional[Mapping[str, Any]] = None
    ) -> Optional[DPModellerConfig]:
        """Converts a dict-based configuration into a DPModellerConfig.

        If the configuration is None, will be returned as-is.
        """
        if dict_config:
            return DPModellerConfig(**dict_config)
        else:
            return None

    def apply_pod_dp(self, pod_dp_config: Optional[DPPodConfig]) -> None:
        """Applies pod-based DP caps and bounds to configuration options.

        Args:
            pod_dp_config: Pod-based configuration related to DP.
        """
        if not self._dp_config:  # DP not being used
            return
        if not pod_dp_config:
            logger.info("No pod DP preferences, using modeller preferences.")
            return

        # Modify maximum epsilon budget based on pod cap
        if self._dp_config.epsilon > pod_dp_config.epsilon:
            logger.warning(
                f"Requested DP max epsilon ({self._dp_config.epsilon}) exceeds "
                f"maximum value allowed by pod. Using pod max of "
                f"{pod_dp_config.epsilon}."
            )
            self._dp_config.epsilon = pod_dp_config.epsilon

        # Modify maximum target delta based on pod cap
        if self._dp_config.delta > pod_dp_config.delta:
            logger.warning(
                f"Requested DP target delta ({self._dp_config.delta}) exceeds "
                f"maximum value allowed by pod. Using pod max of "
                f"{pod_dp_config.delta}."
            )
            self._dp_config.delta = pod_dp_config.delta

    class _Schema(MarshmallowSchema):
        # Two fields are needed to keep the config field private
        _dp_config = fields.Nested(
            desert.schema(DPModellerConfig),
            allow_none=True,
            data_key="dp_config",
            dump_only=True,
        )
        dp_config = fields.Nested(
            desert.schema(DPModellerConfig),
            allow_none=True,
            data_key="dp_config",
            load_only=True,
        )



"""Contains models for running on the Bitfount platform."""
from typing import List

__all__: List[str] = []

# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Tests classes and functions in bitfount/federated/aggregators/base.py."""

import datetime
from typing import cast

import numpy as np
import pandas as pd
import pytest
from pytest import fixture
import torch

from bitfount.federated.aggregators.base import AggregationType, _BaseAggregator
from bitfount.federated.exceptions import AggregatorError
from tests.utils.helper import unit_test


@fixture
def dataframe() -> pd.DataFrame:
    """Returns a dataframe."""
    return pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "c": ["7", "8", "9"]})


@unit_test
class TestBaseAggregator:
    """Tests _BaseAggregator methods."""

    @fixture
    def base_aggregator(self) -> _BaseAggregator:
        """Create _BaseAggregator for tests."""
        return _BaseAggregator()

    def test_get_type_with_mismatching_types(
        self, base_aggregator: _BaseAggregator, dataframe: pd.DataFrame
    ) -> None:
        """Test get_type with mismatching types raises AggregatorError."""
        with pytest.raises(AggregatorError):
            base_aggregator._get_type([dataframe, dataframe["a"].to_numpy()])

    @pytest.mark.parametrize("length", [1, 2, 5])
    def test_get_type_with_numpy_array(
        self, base_aggregator: _BaseAggregator, length: int
    ) -> None:
        """Test get_type with numpy array returns NUMPY_ARRAY."""
        assert (
            base_aggregator._get_type([np.array([1, 2, 3])] * length)
            == AggregationType.NUMPY_ARRAY
        )

    @pytest.mark.parametrize("length", [1, 2, 5])
    def test_get_type_with_pandas_dataframe(
        self, base_aggregator: _BaseAggregator, dataframe: pd.DataFrame, length: int
    ) -> None:
        """Test get_type with pandas dataframe returns PANDAS_DATAFRAME."""
        assert (
            base_aggregator._get_type([dataframe] * length)
            == AggregationType.PANDAS_DATAFRAME
        )

    @pytest.mark.parametrize("length", [1, 2, 5])
    def test_get_type_with_tensor_dictionary(
        self, base_aggregator: _BaseAggregator, length: int
    ) -> None:
        """Test get_type with tensor dictionary returns TENSOR."""
        assert (
            base_aggregator._get_type(
                [{"param1": cast(torch.Tensor, np.array([1, 2, 3]))}] * length
            )
            == AggregationType.TENSOR_DICT
        )

    def test_get_type_with_unsupported_type(
        self, base_aggregator: _BaseAggregator
    ) -> None:
        """Test get_type with unsupported type raises AggregatorError."""
        assert base_aggregator._get_type([object()]) == AggregationType.UNSUPPORTED

    def test_set_non_numeric_columns_with_mismatching_non_numeric_columns(
        self, base_aggregator: _BaseAggregator, dataframe: pd.DataFrame
    ) -> None:
        """Test set_non_numeric_columns with mismatching non_numeric columns."""
        dataframe2 = dataframe.copy(deep=True)
        dataframe2["c"] = ["8", "9", "10"]  # should be 7, 8, 9
        with pytest.raises(
            AggregatorError,
            match="Non-numeric columns must be the same. "
            "Column c has different values between workers.",
        ):
            base_aggregator._set_non_numeric_columns([dataframe, dataframe2])

    def test_set_non_numeric_columns_with_misordered_non_numeric_columns(
        self, base_aggregator: _BaseAggregator, dataframe: pd.DataFrame
    ) -> None:
        """Test set_non_numeric_columns with mis-ordered non_numeric columns.

        This tests that the order of the values in the non-numeric columns is important.
        """
        dataframe2 = dataframe.copy(deep=True)
        dataframe2["c"] = ["7", "9", "8"]  # should be 7, 8, 9
        with pytest.raises(
            AggregatorError,
            match="Non-numeric columns must be the same. "
            "Column c has different values between workers.",
        ):
            base_aggregator._set_non_numeric_columns([dataframe, dataframe2])

    def test_set_non_numeric_columns_sets_correct_columns(
        self, base_aggregator: _BaseAggregator, dataframe: pd.DataFrame
    ) -> None:
        """Test set_non_numeric_columns sets non-numeric columns correctly."""
        dataframe["d"] = [datetime.datetime.today()] * 3
        base_aggregator._set_non_numeric_columns([dataframe, dataframe])
        assert list(base_aggregator.non_numeric_columns) == ["c", "d"]
        np.testing.assert_array_equal(
            base_aggregator.non_numeric_columns["c"], dataframe["c"].to_numpy()
        )
        np.testing.assert_array_equal(
            base_aggregator.non_numeric_columns["d"],
            dataframe["d"].to_numpy(),
        )



"""Pytorch implementation of the tensor shim."""
from __future__ import annotations

from typing import Any, ClassVar, List, Sequence, Union, cast

import numpy as np
import torch

from bitfount.backends.pytorch.types import _AdaptorForPyTorchTensor
from bitfount.federated.shim import BackendTensorShim
from bitfount.types import (
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    _BaseSerializableObjectMixIn,
    _TensorLike,
)


class PyTorchBackendTensorShim(BackendTensorShim, _BaseSerializableObjectMixIn):
    """PyTorch backend shim/bridge for converting from/to PyTorch tensors."""

    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}

    @staticmethod
    def to_numpy(t: Union[_TensorLike, List[float]]) -> np.ndarray:
        """See base class."""
        if isinstance(t, _AdaptorForPyTorchTensor):
            array_t = t.torchtensor.numpy()
        else:
            array_t = np.asarray(t)
        return cast(np.ndarray, array_t)

    @staticmethod
    def to_tensor(p: Sequence, **kwargs: Any) -> _TensorLike:
        """See base class."""
        return _AdaptorForPyTorchTensor(torch.tensor(p, **kwargs))

    @staticmethod
    def to_list(p: Union[np.ndarray, _TensorLike]) -> List[float]:
        """See base class."""
        if isinstance(p, np.ndarray):
            return cast(List[float], p.tolist())
        elif isinstance(p, _AdaptorForPyTorchTensor):
            return p.torchtensor.tolist()
        else:
            raise TypeError("Unexpected type")

    @staticmethod
    def is_tensor(p: Any) -> bool:
        """See base class."""
        is_tensor: bool = torch.is_tensor(p)
        return is_tensor

    @staticmethod
    def clamp_params(
        p: _TensorLike, prime_q: int, precision: int, num_workers: int
    ) -> _TensorLike:
        """Method for clipping params for secure sharing.

        Constrains the parameters for secure sharing to be within the
        required range for secure sharing. Used only when
        `steps_between_parameter_updates` is 1.

        Args:
            p: The tensor to be constrained.
            prime_q: The prime use for secret aggregation.
            precision: The precision used for secret aggregation.
            num_workers: The number of workers taking part in the secure aggregation.

        Returns:
            The clamped parameters.
        """
        if isinstance(p, _AdaptorForPyTorchTensor):
            return _AdaptorForPyTorchTensor(
                torch.clamp(
                    p.torchtensor,
                    -prime_q / (precision * 2 * num_workers),
                    prime_q / (precision * 2 * num_workers),
                )
            )
        else:
            return _AdaptorForPyTorchTensor(
                torch.clamp(
                    p,
                    -prime_q / (precision * 2 * num_workers),
                    prime_q / (precision * 2 * num_workers),
                )
            )
